# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Actor network for Register Allocation."""

from typing import Optional, Sequence, Callable, Text, Any

import gin
import tensorflow as tf
from tf_agents.networks import categorical_projection_network
from tf_agents.networks import encoding_network
from tf_agents.networks import network
from tf_agents.typing import types
from tf_agents.utils import nest_utils


class RegAllocEncodingNetwork(encoding_network.EncodingNetwork):

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    # remove the first layer (Flatten) in postprocessing_layers cause this will
    # flatten the B x T x 33 x dim to B x T x (33 x dim).
    self._postprocessing_layers = self._postprocessing_layers[1:]


class RegAllocProbProjectionNetwork(
    categorical_projection_network.CategoricalProjectionNetwork):

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    # shape after projection_layer: B x T x 33 x 1; then gets re-shaped to
    # B x T x 33.
    self._projection_layer = tf.keras.layers.Dense(
        1,
        kernel_initializer=tf.compat.v1.keras.initializers.VarianceScaling(
            scale=kwargs['logits_init_output_factor']),
        bias_initializer=tf.keras.initializers.Zeros(),
        name='logits')


@gin.configurable
class RegAllocRNDEncodingNetwork(RegAllocEncodingNetwork):

  def __init__(self, **kwargs):
    pooling_layer = tf.keras.layers.GlobalMaxPool1D(data_format='channels_last')
    super().__init__(**kwargs)
    # add a pooling layer at the end to to convert B x T x 33 x dim to
    # B x T x dim.
    self._postprocessing_layers.append(pooling_layer)


@gin.configurable
class RegAllocNetwork(network.DistributionNetwork):
  """Creates the actor network for register allocation policy training."""

  def __init__(
      self,
      input_tensor_spec: types.NestedTensorSpec,
      output_tensor_spec: types.NestedTensorSpec,
      preprocessing_layers: Optional[types.NestedLayer] = None,
      preprocessing_combiner: Optional[tf.keras.layers.Layer] = None,
      conv_layer_params: Optional[Sequence[Any]] = None,
      fc_layer_params: Optional[Sequence[int]] = (200, 100),
      dropout_layer_params: Optional[Sequence[float]] = None,
      activation_fn: Callable[[types.Tensor],
                              types.Tensor] = tf.keras.activations.relu,
      kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
      batch_squash: bool = True,
      dtype: tf.DType = tf.float32,
      name: Text = 'RegAllocNetwork'):
    """Creates an instance of `RegAllocNetwork`.

    Args:
      input_tensor_spec: A nest of `tensor_spec.TensorSpec` representing the
        input.
      output_tensor_spec: A nest of `tensor_spec.BoundedTensorSpec` representing
        the output.
      preprocessing_layers: (Optional.) A nest of `tf.keras.layers.Layer`
        representing preprocessing for the different observations.
        All of these layers must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      preprocessing_combiner: (Optional.) A keras layer that takes a flat list
        of tensors and combines them. Good options include
        `tf.keras.layers.Add` and `tf.keras.layers.Concatenate(axis=-1)`.
        This layer must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      conv_layer_params: Optional list of convolution layers parameters, where
        each item is a length-three tuple indicating (filters, kernel_size,
        stride).
      fc_layer_params: Optional list of fully_connected parameters, where each
        item is the number of units in the layer.
      dropout_layer_params: Optional list of dropout layer parameters, each item
        is the fraction of input units to drop or a dictionary of parameters
        according to the keras.Dropout documentation. The additional parameter
        `permanent`, if set to True, allows to apply dropout at inference for
        approximated Bayesian inference. The dropout layers are interleaved with
        the fully connected layers; there is a dropout layer after each fully
        connected layer, except if the entry in the list is None. This list must
        have the same length of fc_layer_params, or be None.
      activation_fn: Activation function, e.g. tf.nn.relu, slim.leaky_relu, ...
      kernel_initializer: Initializer to use for the kernels of the conv and
        dense layers. If none is provided a default glorot_uniform.
      batch_squash: If True the outer_ranks of the observation are squashed into
        the batch dimension. This allow encoding networks to be used with
        observations with shape [BxTx...].
      dtype: The dtype to use by the convolution and fully connected layers.
      name: A string representing name of the network.

    Raises:
      ValueError: If `input_tensor_spec` contains more than one observation.
    """

    if not kernel_initializer:
      kernel_initializer = tf.compat.v1.keras.initializers.glorot_uniform()

    # input: B x T x obs_spec
    # output: B x T x 33 x dim
    encoder = RegAllocEncodingNetwork(
        input_tensor_spec=input_tensor_spec,
        preprocessing_layers=preprocessing_layers,
        preprocessing_combiner=preprocessing_combiner,
        conv_layer_params=conv_layer_params,
        fc_layer_params=fc_layer_params,
        dropout_layer_params=dropout_layer_params,
        activation_fn=activation_fn,
        kernel_initializer=kernel_initializer,
        batch_squash=batch_squash,
        dtype=dtype)

    projection_network = RegAllocProbProjectionNetwork(
        sample_spec=output_tensor_spec, logits_init_output_factor=0.1)
    output_spec = projection_network.output_spec

    super().__init__(
        input_tensor_spec=input_tensor_spec,
        state_spec=(),
        output_spec=output_spec,
        name=name)

    self._encoder = encoder
    self._projection_network = projection_network
    self._output_tensor_spec = output_tensor_spec

  @property
  def output_tensor_spec(self):
    return self._output_tensor_spec

  def call(self,
           observations: types.NestedTensor,
           step_type: types.NestedTensor,
           network_state=(),
           training: bool = False,
           mask=None):
    _ = mask
    state, network_state = self._encoder(
        observations,
        step_type=step_type,
        network_state=network_state,
        training=training)
    outer_rank = nest_utils.get_outer_rank(observations, self.input_tensor_spec)

    # mask un-evictable registers.
    distribution, _ = self._projection_network(
        state, outer_rank, training=training, mask=observations['mask'])

    return distribution, network_state



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the inlining for size environment."""

import subprocess

import gin
import os

from compiler_opt.rl import env
from compiler_opt.rl.inlining import config

from typing import Dict, List, Optional

_COMPILED_MODULE_NAME = 'compiled_module'


@gin.configurable
class InliningForSizeTask(env.MLGOTask):
  """Implementation of the inlining-for-size MLGOTask."""

  def __init__(self, llvm_size_path: str):
    super().__init__()
    self._llvm_size_path = llvm_size_path

  def get_cmdline(self, clang_path: str, base_args: List[str],
                  interactive_base_path: Optional[str],
                  working_dir: str) -> List[str]:
    if interactive_base_path:
      interactive_args = [
          '-mllvm',
          '-enable-ml-inliner=release',
          '-mllvm',
          f'-inliner-interactive-channel-base={interactive_base_path}',
          #'-mllvm',
          #'-inliner-interactive-include-default',
      ]
    else:
      interactive_args = []
    compiled_module_path = os.path.join(working_dir, _COMPILED_MODULE_NAME)
    return [clang_path
           ] + base_args + interactive_args + ['-o', compiled_module_path]

  def get_module_scores(self, working_dir: str) -> Dict[str, float]:
    compiled_module_path = os.path.join(working_dir, _COMPILED_MODULE_NAME)
    cmdline = [self._llvm_size_path, compiled_module_path]
    completed_proc = subprocess.run(cmdline, capture_output=True, check=True)
    if not completed_proc.stdout:
      raise RuntimeError(f'Empty llvm-size output: {" ".join(cmdline)}')
    output = completed_proc.stdout.decode('utf-8')
    tmp = output.split('\n')
    if len(tmp) != 3:
      raise RuntimeError(f'Wrong llvm-size output {output}')
    tmp = tmp[1].split('\t')
    native_size = int(tmp[0])
    return {'default': native_size}


@gin.configurable
def get_inlining_env(clang_path: str) -> env.MLGOEnvironmentBase:
  time_step_spec, action_spec = config.get_inlining_signature_spec()

  return env.MLGOEnvironmentBase(
      clang_path=clang_path,
      task_type=InliningForSizeTask,
      obs_spec=time_step_spec.observation,
      action_spec=action_spec,
  )



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for local worker manager."""

import concurrent.futures
import time

from absl.testing import absltest
from compiler_opt.distributed.worker import Worker
from compiler_opt.distributed.local import local_worker_manager
from tf_agents.system import system_multiprocessing as multiprocessing


class JobNormal(Worker):
  """Test worker."""

  def __init__(self, arg, *, kwarg):
    self._token = 0
    self._arg = arg
    self._kwarg = kwarg

  @classmethod
  def is_priority_method(cls, method_name: str) -> bool:
    return method_name == 'priority_method'

  def priority_method(self):
    return f'priority {self._token}'

  def get_token(self):
    return self._token

  def set_token(self, value):
    self._token = value

  def get_arg(self):
    return self._arg

  def get_kwarg(self):
    return self._kwarg


class JobFail(Worker):

  def __init__(self, wont_be_passed):
    self._arg = wont_be_passed

  def method(self):
    return self._arg


class JobSlow(Worker):

  def method(self):
    time.sleep(3600)


class LocalWorkerManagerTest(absltest.TestCase):

  def test_pool(self):

    arg = 'foo'
    kwarg = 'bar'

    with local_worker_manager.LocalWorkerPoolManager(
        JobNormal, 2, arg, kwarg=kwarg) as pool:
      p1 = pool.get_currently_active()[0]
      p2 = pool.get_currently_active()[1]
      set_futures = [p1.set_token(1), p2.set_token(2)]
      done, not_done = concurrent.futures.wait(set_futures)
      self.assertLen(done, 2)
      self.assertEmpty(not_done)
      self.assertLen([f for f in done if not f.exception()], 2)
      self.assertEqual(p1.get_token().result(), 1)
      self.assertEqual(p2.get_token().result(), 2)
      self.assertEqual(p1.priority_method().result(), 'priority 1')
      self.assertEqual(p2.priority_method().result(), 'priority 2')
      self.assertEqual(p1.get_arg().result(), 'foo')
      self.assertEqual(p2.get_arg().result(), 'foo')
      self.assertEqual(p2.get_kwarg().result(), 'bar')
      self.assertEqual(p2.get_kwarg().result(), 'bar')
      # wait - to make sure the pump doesn't panic if there's no new messages
      time.sleep(3)
      # everything still works
      self.assertEqual(p2.get_token().result(), 2)

  def test_failure(self):

    with local_worker_manager.LocalWorkerPoolManager(JobFail, 2) as pool:
      with self.assertRaises(concurrent.futures.CancelledError):
        # this will fail because we didn't pass the arg to the ctor, so the
        # worker hosting process will crash.
        pool.get_currently_active()[0].method().result()

  def test_worker_crash_while_waiting(self):

    with local_worker_manager.LocalWorkerPoolManager(JobSlow, 2) as pool:
      p = pool.get_currently_active()[0]
      f = p.method()
      self.assertFalse(f.done())
      try:
        p._process.kill()  # pylint: disable=protected-access
      finally:
        with self.assertRaises(concurrent.futures.CancelledError):
          _ = f.result()


if __name__ == '__main__':
  multiprocessing.handle_test_main(absltest.main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for corpus and related concepts."""
# pylint: disable=protected-access
import os
import re

import tensorflow as tf

from compiler_opt.rl import corpus


class CommandParsingTest(tf.test.TestCase):

  def test_deletion(self):
    self.assertEqual(
        ('-cc1',),
        corpus._apply_cmdline_filters(
            orig_options=('-cc1', '-fthinlto-index=bad', '-split-dwarf-file',
                          '/tmp/foo.dwo', '-split-dwarf-output',
                          'somepath/some.dwo'),
            delete_flags=('-split-dwarf-file', '-split-dwarf-output',
                          '-fthinlto-index', '-fprofile-sample-use',
                          '-fprofile-remapping-file')))

    # OK to not match deletion flags
    self.assertEqual(
        ('-cc1',),
        corpus._apply_cmdline_filters(
            orig_options=('-cc1',),
            delete_flags=('-split-dwarf-file', '-split-dwarf-output',
                          '-fthinlto-index', '-fprofile-sample-use',
                          '-fprofile-remapping-file')))

  def test_addition(self):
    self.assertEqual(
        ('-cc1', '-fix-all-bugs', '-something={context.module_full_path}'),
        corpus._apply_cmdline_filters(
            orig_options=('-cc1',),
            additional_flags=('-fix-all-bugs',
                              '-something={context.module_full_path}')))

  def test_replacement(self):

    # if we expect to be able to replace a flag, and it's not in the original
    # cmdline, raise.
    with self.assertRaises(
        ValueError,
        msg='flags that were expected to be replaced were not found'):
      corpus._apply_cmdline_filters(
          orig_options=('-cc1',),
          replace_flags={'-replace-me': '{context.some_field}.replaced'})

    self.assertEqual(
        ('-cc1', '-replace-me={context.some_field}.replaced'),
        corpus._apply_cmdline_filters(
            orig_options=('-cc1', '-replace-me=some_value'),
            replace_flags={'-replace-me': '{context.some_field}.replaced'}))
    # variant without '='
    self.assertEqual(
        ('-cc1', '-replace-me', '{context.some_field}.replaced'),
        corpus._apply_cmdline_filters(
            orig_options=('-cc1', '-replace-me', 'some_value'),
            replace_flags={'-replace-me': '{context.some_field}.replaced'}))


class ModuleSpecTest(tf.test.TestCase):

  def test_loadable_spec(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='smth', size=1)])
    lms = cps.load_module_spec(cps.module_specs[0])
    corpdir2 = self.create_tempdir()
    fqcmd = lms.build_command_line(corpdir2)
    bc_loc = os.path.join(corpdir2, 'smth', 'input.bc')
    self.assertEqual(('-cc1', '-x', 'ir', f'{bc_loc}'), fqcmd)
    with tf.io.gfile.GFile(bc_loc, 'rb') as f:
      self.assertEqual(f.readlines(), [bytes([1])])


class CorpusTest(tf.test.TestCase):

  def test_constructor(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='1', size=1)],
        additional_flags=('-add',))
    self.assertEqual(cps.module_specs, (corpus.ModuleSpec(
        name='1',
        size=1,
        command_line=('-cc1', '-x', 'ir', '{context.module_full_path}', '-add'),
        has_thinlto=False),))
    self.assertEqual(len(cps), 1)

  def test_invalid_args(self):
    with self.assertRaises(
        ValueError, msg='-cc1 flag not present in .cmd file'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-hi',))

    with self.assertRaises(
        ValueError, msg='do not use add/delete flags to replace'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-cc1',),
          additional_flags=('-fsomething',),
          replace_flags={'-fsomething': 'does not matter'})

    with self.assertRaises(
        ValueError, msg='do not use add/delete flags to replace'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-cc1',),
          additional_flags=('-fsomething=new_value',),
          replace_flags={'-fsomething': 'does not matter'})

    with self.assertRaises(
        ValueError, msg='do not use add/delete flags to replace'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-cc1',),
          delete_flags=('-fsomething',),
          replace_flags={'-fsomething': 'does not matter'})

    with self.assertRaises(
        ValueError, msg='do not use add/delete flags to replace'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-cc1',),
          additional_flags=('-fsomething',),
          delete_flags=('-fsomething',))

    with self.assertRaises(
        ValueError, msg='do not use add/delete flags to replace'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[corpus.ModuleSpec(name='smol', size=1)],
          cmdline=('-cc1',),
          additional_flags=('-fsomething=value',),
          delete_flags=('-fsomething',))

  def test_empty_module_list(self):
    location = self.create_tempdir()
    with self.assertRaises(
        ValueError,
        msg=f'{location}\'s corpus_description contains no modules.'):
      corpus.create_corpus_for_testing(
          location=self.create_tempdir(), elements=[], cmdline=('-hello',))

  def test_ctor_thinlto(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='smol', size=1)],
        cmdline=('-cc1', '-fthinlto-index=foo'),
        is_thinlto=True)
    self.assertIn('-fthinlto-index={context.thinlto_full_path}',
                  cps.module_specs[0].command_line)
    self.assertEqual(cps.module_specs[0].command_line[-5:],
                     ('-x', 'ir', '{context.module_full_path}', '-mllvm',
                      '-thinlto-assume-merged'))

  def test_braces_in_cmd(self):
    corpusdir = self.create_tempdir()
    cps = corpus.create_corpus_for_testing(
        location=corpusdir,
        elements=[corpus.ModuleSpec(name='somename', size=1)],
        cmdline=('-cc1', r'-DMACRO(expr)=do {} while(0)'),
        additional_flags=('-additional_flag={context.module_full_path}',))
    mod_spec = cps.module_specs[0]
    loaded_spec = cps.load_module_spec(mod_spec)
    corpdir2 = self.create_tempdir()
    final_cmdline = loaded_spec.build_command_line(corpdir2)
    bcpath = os.path.join(corpdir2, 'somename/input.bc')
    self.assertEqual(final_cmdline,
                     ('-cc1', r'-DMACRO(expr)=do {} while(0)', '-x', 'ir',
                      bcpath, f'-additional_flag={bcpath}'))

  def test_cmd_override_thinlto(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='smol', size=1)],
        cmdline=(),
        cmdline_is_override=True,
        is_thinlto=True)
    self.assertNotIn('-fthinlto-index', cps.module_specs[0].command_line)
    self.assertEqual(cps.module_specs[0].command_line[-6:],
                     ('-x', 'ir', '{context.module_full_path}',
                      '-fthinlto-index={context.thinlto_full_path}', '-mllvm',
                      '-thinlto-assume-merged'))

    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='smol', size=1)],
        cmdline=('-something',),
        cmdline_is_override=True,
        is_thinlto=True)
    self.assertIn('-fthinlto-index={context.thinlto_full_path}',
                  cps.module_specs[0].command_line)
    self.assertEqual(cps.module_specs[0].command_line[-6:],
                     ('-x', 'ir', '{context.module_full_path}',
                      '-fthinlto-index={context.thinlto_full_path}', '-mllvm',
                      '-thinlto-assume-merged'))
    self.assertIn('-something', cps.module_specs[0].command_line)

  def test_sample(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[
            corpus.ModuleSpec(name='smol', size=1),
            corpus.ModuleSpec(name='middle', size=200),
            corpus.ModuleSpec(name='largest', size=500),
            corpus.ModuleSpec(name='small', size=100)
        ])
    sample = cps.sample(4, sort=True)
    self.assertLen(sample, 4)
    self.assertEqual(sample[0].name, 'largest')
    self.assertEqual(sample[1].name, 'middle')
    self.assertEqual(sample[2].name, 'small')
    self.assertEqual(sample[3].name, 'smol')

  def test_sample_without_replacement(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[
            corpus.ModuleSpec(name='smol', size=1),
            corpus.ModuleSpec(name='middle', size=200),
            corpus.ModuleSpec(name='largest', size=500),
            corpus.ModuleSpec(name='small', size=100)
        ],
        sampler_type=corpus.SamplerWithoutReplacement)
    samples = []
    samples.extend(cps.sample(1, sort=True))
    self.assertLen(samples, 1)
    samples.extend(cps.sample(1, sort=True))
    self.assertLen(samples, 2)
    # Can't sample 3 from the corpus because there are only 2 elements left
    with self.assertRaises(corpus.CorpusExhaustedError):
      samples.extend(cps.sample(3, sort=True))
    # But, we can sample exactly 2 more
    self.assertLen(samples, 2)
    samples.extend(cps.sample(2, sort=True))
    self.assertLen(samples, 4)
    with self.assertRaises(corpus.CorpusExhaustedError):
      samples.extend(cps.sample(1, sort=True))
    samples.sort(key=lambda m: m.size, reverse=True)
    self.assertEqual(samples[0].name, 'largest')
    self.assertEqual(samples[1].name, 'middle')
    self.assertEqual(samples[2].name, 'small')
    self.assertEqual(samples[3].name, 'smol')

  def test_filter(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[
            corpus.ModuleSpec(name='smol', size=1),
            corpus.ModuleSpec(name='middle', size=200),
            corpus.ModuleSpec(name='largest', size=500),
            corpus.ModuleSpec(name='small', size=100)
        ],
        module_filter=lambda name: re.compile(r'.+l').match(name))
    sample = cps.sample(999, sort=True)
    self.assertLen(sample, 3)
    self.assertEqual(sample[0].name, 'middle')
    self.assertEqual(sample[1].name, 'small')
    self.assertEqual(sample[2].name, 'smol')

  def test_sample_zero(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name='smol', size=1)])

    self.assertRaises(ValueError, cps.sample, 0)
    self.assertRaises(ValueError, cps.sample, -213213213)

  def test_bucket_sample(self):
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name=f'{i}', size=i) for i in range(100)])
    # Odds of passing once by pure luck with random.sample: 1.779e-07
    # Try 32 times, for good measure.
    for i in range(32):
      sample = cps.sample(k=20, sort=True)
      self.assertLen(sample, 20)
      for idx, s in enumerate(sample):
        # Each bucket should be size 5, since n=20 in the sampler
        self.assertEqual(s.size // 5, 19 - idx)

  def test_bucket_sample_all(self):
    # Make sure we can sample everything, even if it's not divisible by the
    # `n` in SamplerBucketRoundRobin.
    # Create corpus with a prime number of modules.
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name=f'{i}', size=i) for i in range(101)])

    # Try 32 times, for good measure.
    for i in range(32):
      sample = cps.sample(k=101, sort=True)
      self.assertLen(sample, 101)
      for idx, s in enumerate(sample):
        # Since everything is sampled, it should be in perfect order.
        self.assertEqual(s.size, 100 - idx)

  def test_bucket_sample_small(self):
    # Make sure we can sample even when k < n.
    cps = corpus.create_corpus_for_testing(
        location=self.create_tempdir(),
        elements=[corpus.ModuleSpec(name=f'{i}', size=i) for i in range(100)])

    # Try all 19 possible values 0 < i < n
    for i in range(1, 20):
      sample = cps.sample(k=i, sort=True)
      self.assertLen(sample, i)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A tool for analyzing which features a model uses to make a decision.

This script allows for processing a set of examples generated from a trace
created through generate_default_trace into a set of shap values which
represent how much that specific feature contributes to the final output of
the model. These values can then be imported into an IPython notebook and
graphed with the help of the feature_importance_graphs.py module in the same
folder.

Usage:
PYTHONPATH=$PYTHONPATH:. python3 compiler_opt/tools/feature_importance.py \
    --gin_files=compiler_opt/rl/regalloc/gin_configs/common.gin \
    --gin_bindings=config_registry.get_configuration.implementation=\
      @configs.RegallocEvictionConfig \
    --data_path=/default_trace \
    --model_path=/warmstart/saved_policy \
    --num_examples=5 \
    --output_file=./explanation_data.json

The type of trace that is performed (ie if it is just tracing the default
heuristic or if it is a trace of a ML model) doesn't matter as the only data
that matters re the input features. The num_examples flag sets the number of
examples that get processed into shap values. Increasing this value will
potentially allow you to reach better conclusions depending upon how you're
viewing the data, but increasing it will also increase the runtime of this
script quite significantly as the process is not multithreaded.
"""

from absl import app
from absl import flags
from absl import logging
import gin

from compiler_opt.rl import agent_config
from compiler_opt.rl import data_reader
from compiler_opt.rl import registry

from compiler_opt.tools import feature_importance_utils

import tensorflow as tf
import shap
import numpy
import numpy.typing
import json

_DATA_PATH = flags.DEFINE_multi_string(
    'data_path', [], 'Path to TFRecord file(s) containing trace data.')
_MODEL_PATH = flags.DEFINE_string('model_path', '',
                                  'Path to the model to explain')
_OUTPUT_FILE = flags.DEFINE_string(
    'output_file', '', 'The path to the output file containing the SHAP values')
_NUM_EXAMPLES = flags.DEFINE_integer(
    'num_examples', 1, 'The number of examples to process from the trace')
_GIN_FILES = flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files.')
_GIN_BINDINGS = flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')


def main(_):
  gin.parse_config_files_and_bindings(
      _GIN_FILES.value, bindings=_GIN_BINDINGS.value, skip_unknown=False)
  logging.info(gin.config_str())

  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  agent_cfg = agent_config.BCAgentConfig(
      time_step_spec=time_step_spec, action_spec=action_spec)
  tfrecord_dataset_fn = data_reader.create_tfrecord_dataset_fn(
      agent_cfg=agent_cfg, batch_size=1, train_sequence_length=1)

  dataset_iter = iter(tfrecord_dataset_fn(_DATA_PATH.value).repeat())

  raw_trajectory = next(dataset_iter)

  saved_policy = tf.saved_model.load(_MODEL_PATH.value)
  action_fn = saved_policy.signatures['action']

  observation = feature_importance_utils.process_raw_trajectory(raw_trajectory)
  input_sig = feature_importance_utils.get_input_signature(observation)

  run_model = feature_importance_utils.create_run_model_function(
      action_fn, input_sig)

  total_size = feature_importance_utils.get_signature_total_size(input_sig)
  flattened_input = feature_importance_utils.flatten_input(
      observation, total_size)
  flattened_input = numpy.expand_dims(flattened_input, axis=0)
  dataset = numpy.empty((_NUM_EXAMPLES.value, total_size))
  for i in range(0, _NUM_EXAMPLES.value):
    raw_trajectory = next(dataset_iter)
    observation = feature_importance_utils.process_raw_trajectory(
        raw_trajectory)
    flat_input = feature_importance_utils.flatten_input(observation, total_size)
    dataset[i] = flat_input

  explainer = shap.KernelExplainer(run_model, numpy.zeros((1, total_size)))
  shap_values = explainer.shap_values(dataset, nsamples=1000)
  processed_shap_values = feature_importance_utils.collapse_values(
      input_sig, shap_values, _NUM_EXAMPLES.value)

  # if we have more than one value per feature, just set the dataset to zeros
  # as summing across a dimension produces data that doesn't really mean
  # anything
  if feature_importance_utils.get_max_part_size(input_sig) > 1:
    dataset = numpy.zeros(processed_shap_values.shape)

  feature_names = list(input_sig.keys())

  output_file_data = {
      'expected_values': explainer.expected_value,
      'shap_values': processed_shap_values.tolist(),
      'data': dataset.tolist(),
      'feature_names': feature_names
  }

  with open(_OUTPUT_FILE.value, 'w', encoding='utf-8') as output_file:
    json.dump(output_file_data, output_file)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""util function to create training datasets."""

from typing import Callable, List

import tensorflow as tf
from tf_agents.trajectories import trajectory

from compiler_opt.rl import agent_config


def create_parser_fn(
    agent_cfg: agent_config.AgentConfig
) -> Callable[[str], trajectory.Trajectory]:
  """Create a parser function for reading from a serialized tf.SequenceExample.

  Args:
    agent_name: AgentName, enum type of the agent.
    time_step_spec: time step spec of the optimization problem.
    action_spec: action spec of the optimization problem.

  Returns:
    A callable that takes scalar serialized proto Tensors and emits
    `Trajectory` objects containing parsed tensors.
  """

  def _parser_fn(serialized_proto):
    """Helper function that is returned by create_`parser_fn`."""
    # We copy through all context features at each frame, so even though we know
    # they don't change from frame to frame, they are still sequence features
    # and stored in the feature list.
    context_features = {}
    # pylint: disable=g-complex-comprehension
    sequence_features = dict(
        (tensor_spec.name,
         tf.io.FixedLenSequenceFeature(
             shape=tensor_spec.shape, dtype=tensor_spec.dtype))
        for tensor_spec in agent_cfg.time_step_spec.observation.values())
    sequence_features[
        agent_cfg.action_spec.name] = tf.io.FixedLenSequenceFeature(
            shape=agent_cfg.action_spec.shape,
            dtype=agent_cfg.action_spec.dtype)
    sequence_features[
        agent_cfg.time_step_spec.reward.name] = tf.io.FixedLenSequenceFeature(
            shape=agent_cfg.time_step_spec.reward.shape,
            dtype=agent_cfg.time_step_spec.reward.dtype)
    sequence_features.update(agent_cfg.get_policy_info_parsing_dict())

    # pylint: enable=g-complex-comprehension
    with tf.name_scope('parse'):
      _, parsed_sequence = tf.io.parse_single_sequence_example(
          serialized_proto,
          context_features=context_features,
          sequence_features=sequence_features)
      # TODO(yundi): make the transformed reward configurable.
      action = parsed_sequence[agent_cfg.action_spec.name]
      reward = tf.cast(parsed_sequence[agent_cfg.time_step_spec.reward.name],
                       tf.float32)

      policy_info = agent_cfg.process_parsed_sequence_and_get_policy_info(
          parsed_sequence)

      del parsed_sequence[agent_cfg.time_step_spec.reward.name]
      del parsed_sequence[agent_cfg.action_spec.name]
      full_trajectory = trajectory.from_episode(
          observation=parsed_sequence,
          action=action,
          policy_info=policy_info,
          reward=reward)
      return full_trajectory

  return _parser_fn


def create_flat_sequence_example_dataset_fn(
    agent_cfg: agent_config.AgentConfig
) -> Callable[[List[str]], tf.data.Dataset]:
  """Get a function that creates a dataset from serialized sequence examples.

  The dataset is "flat" insofar as it does not batch for sequence length nor
  batches.

  Args:
    agent_name: AgentName, enum type of the agent.
    time_step_spec: time step spec of the optimization problem.
    action_spec: action spec of the optimization problem.

  Returns:
    A callable that takes a list of serialized sequence examples and returns
      a `tf.data.Dataset`.  Treating this dataset as an iterator yields batched
      `trajectory.Trajectory` instances with shape `[...]`.
  """
  parser_fn = create_parser_fn(agent_cfg)

  def _sequence_example_dataset_fn(sequence_examples):
    # Data collector returns empty strings for corner cases, filter them out
    # here.
    # yapf: disable - Looks better hand formatted
    dataset = (tf.data.Dataset
                .from_tensor_slices(sequence_examples)
                .filter(lambda string: tf.strings.length(string) > 0)
                .map(parser_fn)
                .filter(lambda traj: tf.size(traj.reward) > 2)
                .unbatch()
               )
    # yapf: enable
    return dataset

  return _sequence_example_dataset_fn


def create_sequence_example_dataset_fn(
    agent_cfg: agent_config.AgentConfig, batch_size: int,
    train_sequence_length: int) -> Callable[[List[str]], tf.data.Dataset]:
  """Get a function that creates a dataset from serialized sequence examples.

  Args:
    agent_name: AgentName, enum type of the agent.
    time_step_spec: time step spec of the optimization problem.
    action_spec: action spec of the optimization problem.
    batch_size: int, batch_size B.
    train_sequence_length: int, trajectory sequence length T.

  Returns:
    A callable that takes a list of serialized sequence examples and returns
      a `tf.data.Dataset`.  Treating this dataset as an iterator yields batched
      `trajectory.Trajectory` instances with shape `[B, T, ...]`.
  """
  trajectory_shuffle_buffer_size = 1024

  flat_sequence_example_dataset_fn = create_flat_sequence_example_dataset_fn(
      agent_cfg)

  def _sequence_example_dataset_fn(sequence_examples):
    # Data collector returns empty strings for corner cases, filter them out
    # here.
    # yapf: disable - Looks better hand formatted
    dataset = flat_sequence_example_dataset_fn(sequence_examples)
    return (dataset.batch(train_sequence_length, drop_remainder=True)
                   .cache()
                   .shuffle(trajectory_shuffle_buffer_size)
                   .batch(batch_size, drop_remainder=True))

  return _sequence_example_dataset_fn


# TODO(yundi): PyType check of input_dataset as Type[tf.data.Dataset] is not
# working.
def create_file_dataset_fn(
    agent_cfg: agent_config.AgentConfig,
    batch_size: int,
    train_sequence_length: int,
    input_dataset) -> Callable[[List[str]], tf.data.Dataset]:
  """Get a function that creates an dataset from files.

  Args:
    agent_name: AgentName, enum type of the agent.
    time_step_spec: time step spec of the optimization problem.
    action_spec: action spec of the optimization problem.
    batch_size: int, batch_size B.
    train_sequence_length: int, trajectory sequence length T.
    input_dataset: A tf.data.Dataset subclass object.

  Returns:
    A callable that takes file path(s) and returns a `tf.data.Dataset`.
      Iterating over this dataset yields `trajectory.Trajectory` instances with
      shape `[B, T, ...]`.
  """
  files_buffer_size = 100
  num_readers = 10
  num_map_threads = 8
  shuffle_buffer_size = 1024
  trajectory_shuffle_buffer_size = 1024

  parser_fn = create_parser_fn(agent_cfg)

  def _file_dataset_fn(data_path):
    dataset = (
        tf.data.Dataset.list_files(data_path).shuffle(
            files_buffer_size).interleave(
                input_dataset, cycle_length=num_readers, block_length=1)
        # Due to a bug in collection, we sometimes get empty rows.
        .filter(lambda string: tf.strings.length(string) > 0).apply(
            tf.data.experimental.shuffle_and_repeat(shuffle_buffer_size)).map(
                parser_fn, num_parallel_calls=num_map_threads)
        # Only keep sequences of length 2 or more.
        .filter(lambda traj: tf.size(traj.reward) > 2))

    # TODO(yundi): window and subsample data.
    # TODO(yundi): verify the shuffling is correct.
    dataset = (
        dataset.unbatch().batch(
            train_sequence_length,
            drop_remainder=True).shuffle(trajectory_shuffle_buffer_size).batch(
                batch_size,
                drop_remainder=True).prefetch(tf.data.experimental.AUTOTUNE))
    return dataset

  return _file_dataset_fn


def create_tfrecord_dataset_fn(
    agent_cfg: agent_config.AgentConfig, batch_size: int,
    train_sequence_length: int) -> Callable[[List[str]], tf.data.Dataset]:
  """Get a function that creates an dataset from tfrecord.

  Args:
    agent_name: AgentName, enum type of the agent.
    time_step_spec: time step spec of the optimization problem.
    action_spec: action spec of the optimization problem.
    batch_size: int, batch_size B.
    train_sequence_length: int, trajectory sequence length T.

  Returns:
    A callable that takes tfrecord path(s) and returns a `tf.data.Dataset`.
      Iterating over this dataset yields `trajectory.Trajectory` instances with
      shape `[B, T, ...]`.
  """
  return create_file_dataset_fn(
      agent_cfg,
      batch_size,
      train_sequence_length,
      input_dataset=tf.data.TFRecordDataset)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Random Network Distillation Implementation."""
import gin
import tensorflow as tf
from tf_agents.utils import tensor_normalizer


@gin.configurable
class RandomNetworkDistillation():
  """The Random Network Distillation class."""

  def __init__(self,
               time_step_spec=None,
               preprocessing_layer_creator=None,
               encoding_network=None,
               learning_rate=1e-4,
               update_frequency=4,
               fc_layer_params=(32,),
               initial_intrinsic_reward_scale=1.0,
               half_decay_steps=10000):
    """Initilization for RandomNetworkDistillation class.

    Args:
      time_step_spec: the time step spec for raw observation
      preprocessing_layer_creator: A callable returns feature processing layer
        given observation_spec.
      encoding_network: A tf_agents.networks.Network class.
      learning_rate: the learning rate for optimizer.
      update_frequency: the update frequency for the predictor network.
      fc_layer_params: list of fully_connected parameters, where each item is
        the number of units in the layer.
      initial_intrinsic_reward_scale: the scale of the initial intrinsic reward.
      half_decay_steps: the steps for intrinsic reward scale to decay by half.
    """

    feature_extractor_layer = tf.nest.map_structure(preprocessing_layer_creator,
                                                    time_step_spec.observation)

    self._target_net = encoding_network(
        input_tensor_spec=time_step_spec.observation,
        preprocessing_layers=feature_extractor_layer,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=fc_layer_params,
        name='ObsNormalizationNetwork')

    self._predict_net = encoding_network(
        input_tensor_spec=time_step_spec.observation,
        preprocessing_layers=feature_extractor_layer,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=fc_layer_params,
        name='ObsNormalizationNetwork')

    self._predict_net_optimizer = tf.keras.optimizers.Adam(
        learning_rate=learning_rate)
    self._intrinsic_reward_normalizer = (
        tensor_normalizer.StreamingTensorNormalizer(
            tf.TensorSpec([], tf.float32)))
    self._update_frequency = update_frequency
    self._decay_scale = tf.keras.optimizers.schedules.ExponentialDecay(
        initial_intrinsic_reward_scale,
        decay_steps=half_decay_steps,
        decay_rate=0.5,
        staircase=False)

    self._global_step = tf.compat.v1.train.get_or_create_global_step()
    self._normalized_intrinsic_reward_mean = tf.keras.metrics.Mean()
    self._intrinsic_reward_mean = tf.keras.metrics.Mean()
    self._external_reward_mean = tf.keras.metrics.Mean()

  def _get_intrinsic_reward(self, observation):
    """Compute the intrisic reward.

    Args:
      observation: raw observation in observation_spec format

    Returns:
      intrinsic_reward: the intrinsic reward
    """
    with tf.GradientTape() as tape:
      # make the predict network parameters trainable
      # Compute the feature embedding loss (for next obseravtion trajectory)
      feature_target, _ = self._target_net(observation)
      feature_predict, _ = self._predict_net(observation)
      feature_target = tf.stop_gradient(feature_target)

      # compute the embedding loss on a portion of the batch data
      # _update_frequency denotes the stride slicing
      emb_loss = tf.reduce_mean(
          tf.reduce_sum(
              tf.math.square(feature_target[::self._update_frequency] -
                             feature_predict[::self._update_frequency]),
              axis=-1))

      # log the predictor loss
      with tf.name_scope('Losses/'):
        tf.summary.scalar(
            name='random_net_predictor_loss',
            data=emb_loss,
            step=self._global_step)

      # compute the gradient and optimize the predictor function
      pred_grad = tape.gradient(emb_loss, self._predict_net.trainable_variables)
      self._predict_net_optimizer.apply_gradients(
          zip(pred_grad, self._predict_net.trainable_variables))

    # compute the intrinsic reward using the l2 norm square difference
    # only consider the next state trajectory, whose length = original length-1
    intrinsic_reward = tf.reduce_sum(
        tf.math.square(feature_target - feature_predict), axis=-1)[:, 1:]

    return intrinsic_reward

  def _update_metrics(self, experience, intrinsic_reward,
                      normalized_intrinsic_reward):
    """Updates metrics and exports to Tensorboard.

    Args:
      experience: the experience trajectory in shape of `[batch_size,
        time_steps]`.
      intrinsic_reward: the intrinsic reward in shape of `[batch_size,
        time_steps - 1]` (intrinsic_reward is based on the 'next-state'
        trajectory).
      normalized_intrinsic_reward: the scaled normalized intrinsic reward in
        shape of `[batch_size, time_steps - 1]`.
    """
    is_action = ~experience.is_boundary()
    # normalized_intrinsic_reward will assign for the first time_steps - 1
    # length trajectory states. The sample_weight is also based on the first
    # time_steps - 1 length trajectory.
    self._normalized_intrinsic_reward_mean.update_state(
        normalized_intrinsic_reward, sample_weight=is_action[:, :-1])
    self._intrinsic_reward_mean.update_state(
        intrinsic_reward, sample_weight=is_action[:, :-1])
    self._external_reward_mean.update_state(
        experience.reward, sample_weight=is_action)
    with tf.name_scope('random_network_distillation/'):
      tf.summary.scalar(
          name='data_normalized_intrinsic_reward_mean',
          data=self._normalized_intrinsic_reward_mean.result(),
          step=self._global_step)
      tf.summary.scalar(
          name='data_intrinsic_reward_mean',
          data=self._intrinsic_reward_mean.result(),
          step=self._global_step)
      tf.summary.scalar(
          name='data_external_reward_mean',
          data=self._external_reward_mean.result(),
          step=self._global_step)
      tf.summary.scalar(
          name='intrinsic_reward_scale',
          data=self._decay_scale(self._global_step),
          step=self._global_step)

    tf.summary.histogram(
        name='external_reward', data=experience.reward, step=self._global_step)
    tf.summary.histogram(
        name='intrinsic_reward', data=intrinsic_reward, step=self._global_step)
    tf.summary.histogram(
        name='normalized_intrinsic_reward',
        data=normalized_intrinsic_reward,
        step=self._global_step)

  def train(self, experience):
    """Train the predictor on the batched next state trajectory.

    Args:
      experience: Trajectory

    Returns:
      expereince_new: new Trajectory (the new trajectory modified from the
      original experience trajectory, where the reward is updated as the
      addition of external reward + intrinsic reward).
    """
    # compute intrinsic reward for length - 1 horizon
    intrinsic_reward = self._get_intrinsic_reward(experience.observation)

    normalized_intrinsic_reward = self._intrinsic_reward_normalizer.normalize(
        intrinsic_reward, clip_value=0, center_mean=False) * self._decay_scale(
            self._global_step)
    self._intrinsic_reward_normalizer.update(intrinsic_reward)

    # update the log
    self._update_metrics(experience, intrinsic_reward,
                         normalized_intrinsic_reward)

    batch_size = experience.reward.shape[0]
    # assign the last time step reward = 0 (no intrinsic reward)
    # pylint: disable=unexpected-keyword-arg
    normalized_intrinsic_reward = tf.concat(
        [normalized_intrinsic_reward,
         tf.zeros([batch_size, 1])], axis=1)

    # reconstruct the reward: external + intrinsic
    reconstructed_reward = experience.reward + normalized_intrinsic_reward

    return experience.replace(reward=reconstructed_reward)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Analysis for benchmark results.json."""

import collections
import math
import statistics

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from absl import logging

# For each benchmark, and for each counter, capture the recorded values.
PerBenchmarkResults = Dict[str, Dict[str, List[float]]]

# Benchmark data, as captured by the benchmark json output: a dictionary from
# benchmark names to a list of run results. Each run result is a dictionary of
# key-value pairs, e.g. counter name - value.
BenchmarkRunResults = Dict[str, List[Dict[str, Any]]]

# A comparison per benchmark, per counter, capturing the geomean and the stdev
# of the base and experiment values.
ABComparison = Dict[str, Dict[str, Tuple[float, float, float]]]


def _geomean(data: List[float]):
  return math.exp(sum(math.log(x) for x in data) / len(data))


def _stdev(data: List[float]):
  assert data
  return 0.0 if len(data) == 1 else statistics.stdev(data)


class BenchmarkReport:
  """The counter values collected for benchmarks in a benchmark suite."""

  def __init__(self, suite_name: str, json_data: BenchmarkRunResults,
               counter_names: Iterable[str]):
    self._suite_name = suite_name
    self._load_values(json_data, counter_names)

  def suite_name(self):
    return self._suite_name

  def values(self):
    return self._values

  def names(self):
    return self._names

  def counters(self):
    return self._counters

  def raw_measurements(self):
    return self._raw_measurements

  def counter_means(self, benchmark: str, counter: str) -> Tuple[float, float]:
    if counter not in self.counters():
      raise ValueError('unknown counter')
    if benchmark not in self.names():
      raise ValueError('unknown benchmark')
    return (_geomean(self._values[benchmark][counter]),
            _stdev(self._values[benchmark][counter]))

  def zero_counters(self):
    ret = set()
    for name in self.names():
      for counter in self.values()[name]:
        if 0.0 in self.values()[name][counter]:
          ret.add((name, counter))
    return frozenset(ret)

  def large_variation_counters(self, variation: float):
    ret = set()
    for name in self.names():
      for counter in self.values()[name]:
        vals = self.values()[name][counter]
        swing = _stdev(vals) / _geomean(vals)
        if swing > variation:
          ret.add((name, counter, swing))
    return frozenset(ret)

  def _load_values(self, data: BenchmarkRunResults,
                   names: Iterable[str]) -> None:
    """Organize json values per-benchmark, per counter.

    Args:
        data: json data
        names: perf counter names
    Returns:
        benchmark data organized per-benchmark, per-counter name.
    """
    runs = data['benchmarks']
    self._values = collections.defaultdict(
        lambda: collections.defaultdict(list))
    self._raw_measurements = collections.defaultdict(
        lambda: collections.defaultdict(list))
    self._counters = set()
    self._names = set()

    for r in runs:
      benchmark_name = r['name']
      for counter in names:
        value = float(r[counter])
        iters = float(r['iterations'])
        self._raw_measurements[benchmark_name][counter].append(value * iters)
        self._values[benchmark_name][counter].append(value)
        self._counters.add(counter)
        self._names.add(benchmark_name)
    self._counters = frozenset(self._counters)
    self._names = frozenset(self._names)


class BenchmarkComparison:
  """Analysis of 2 benchmark runs."""

  def __init__(self, base_report: BenchmarkReport, exp_report: BenchmarkReport):
    base_names_set = set(base_report.names())
    exp_names_set = set(exp_report.names())
    if base_report.suite_name() != exp_report.suite_name():
      raise ValueError('cannot compare different suites')
    if base_names_set != exp_names_set:
      diff_base_exp = base_names_set.difference(exp_names_set)
      diff_exp_base = exp_names_set.difference(base_names_set)
      diff_set = diff_base_exp.union(diff_exp_base)
      logging.info('The following tests differ between the test suites: %s',
                   diff_set)
      raise ValueError('suite runs have different benchmark names')
    if set(base_report.counters()) != set(exp_report.counters()):
      raise ValueError(
          'counter names are different between base and experiment')

    self._base = base_report
    self._exp = exp_report

  def suite_name(self):
    return self._base.suite_name()

  def summarize(self) -> ABComparison:
    """Summarize the results from two runs (base/experiment).

    Returns:
      A per benchmark, per counter summary of the improvement/regression
      between the 2 runs, in percents.
    """
    base_results = self._base.values()
    exp_results = self._exp.values()

    ret = {}
    for bname in base_results:
      ret[bname] = {}
      for counter in base_results[bname]:
        base_vals = base_results[bname][counter]
        exp_vals = exp_results[bname][counter]
        base_geomean = _geomean(base_vals)
        exp_geomean = _geomean(exp_vals)
        improvement = 1 - exp_geomean / base_geomean
        base_stdev = _stdev(base_vals)
        exp_stdev = _stdev(exp_vals)
        ret[bname][counter] = (improvement, base_stdev / base_geomean,
                               exp_stdev / exp_geomean)
    return ret

  def names(self):
    return self._base.names()

  def counters(self):
    return self._base.counters()

  def total_improvement(self, counter: str):
    assert counter in self.counters()
    logsum = 0
    # we look at the geomean of the improvement for each benchmark
    for bname in self.names():
      b_geomean, _ = self._base.counter_means(bname, counter)
      e_geomean, _ = self._exp.counter_means(bname, counter)
      logsum += math.log(e_geomean / b_geomean)
    return 1.0 - math.exp(logsum / len(self.names()))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.constant_value_network."""

import tensorflow as tf

from tf_agents.specs import tensor_spec

from compiler_opt.rl import constant_value_network


class ConstantValueNetworkTest(tf.test.TestCase):

  def testBuilds(self):
    observation_spec = tensor_spec.BoundedTensorSpec((8, 8, 3), tf.float32, 0,
                                                     1)
    observation = tensor_spec.sample_spec_nest(
        observation_spec, outer_dims=(1,))

    net = constant_value_network.ConstantValueNetwork(
        input_tensor_spec=observation_spec, constant_output_val=0)

    value, _ = net(observation)
    self.assertAllEqual([0.], value)

  def testHandlesExtraOuterDims(self):
    observation_spec = tensor_spec.BoundedTensorSpec((8, 8, 3), tf.float32, 0,
                                                     1)
    observation = tensor_spec.sample_spec_nest(
        observation_spec, outer_dims=(2, 2))

    net = constant_value_network.ConstantValueNetwork(
        input_tensor_spec=observation_spec, constant_output_val=1)

    value, _ = net(observation)
    self.assertAllEqual([[1., 1.], [1., 1.]], value)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common abstraction for a worker contract."""

import abc
import sys
from typing import Any, List, Iterable, Optional, Protocol, TypeVar

import gin


class Worker(Protocol):

  @classmethod
  def is_priority_method(cls, method_name: str) -> bool:
    _ = method_name
    return False


T = TypeVar('T')


class WorkerPool(metaclass=abc.ABCMeta):
  """Abstraction of a pool of workers that may be refreshed."""

  # Issue #155 would strongly-type the return type.
  @abc.abstractmethod
  def get_currently_active(self) -> List[Any]:
    raise NotImplementedError()

  @abc.abstractmethod
  def get_worker_concurrency(self) -> int:
    raise NotImplementedError()


class FixedWorkerPool(WorkerPool):
  """A WorkerPool built from a fixed list of workers."""

  # Issue #155 would strongly-type `workers`
  def __init__(self, workers: List[Any], worker_concurrency: int = 2):
    self._workers = workers
    self._worker_concurrency = worker_concurrency

  def get_currently_active(self):
    return self._workers

  def get_worker_concurrency(self):
    return self._worker_concurrency


# Dask's Futures are limited. This captures that.
class WorkerFuture(Protocol[T]):

  def result(self) -> T:
    raise NotImplementedError()

  def done(self) -> bool:
    raise NotImplementedError()

  def add_done_callback(self, fn) -> None:
    raise NotImplementedError


def wait_for(futures: Iterable[WorkerFuture]):
  """Dask futures don't support more than result() and done()."""
  for f in futures:
    try:
      _ = f.result()
    except:  # pylint: disable=bare-except
      pass


def get_exception(worker_future: WorkerFuture) -> Optional[Exception]:
  assert worker_future.done()
  try:
    _ = worker_future.result()
    return None
  except Exception as e:  # pylint: disable=broad-except
    return e


def get_full_worker_args(worker_class: 'type[Worker]', **current_kwargs):
  """Get the union of given kwargs and gin config.

  This allows the worker hosting process be set up differently from the training
  process - e.g. no need to initialize gin variables there, for example.
  """
  gin_config = {}
  try:
    gin_config = gin.get_bindings(worker_class)
  except ValueError:
    # we don't have a way to check if `worker_class` is even known to gin, and
    # it's not a requirement that it were. Tests, for instance, don't use gin.
    pass
  # Issue #38
  if sys.version_info.minor >= 9:
    return current_kwargs | gin_config
  else:
    return {**current_kwargs, **gin_config}



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for generate_test_model.

An integration test for model saving, to detect TFLite model conversion.
"""

from absl import flags
from absl.testing import absltest
from absl.testing import flagsaver
from absl.testing import parameterized
import gin

from compiler_opt.tools import generate_test_model

flags.FLAGS['gin_files'].allow_override = True
flags.FLAGS['gin_bindings'].allow_override = True


def _get_test_settings():
  test_setting = []

  agent_config_type_dict = {
      'ppo': '@agents.PPOAgentConfig',
      'behavioral_cloning': '@agents.BCAgentConfig',
  }

  for problem in ('inlining', 'regalloc'):
    for algorithm in ('ppo', 'behavioral_cloning'):
      test_name = f'{problem}_{algorithm}'
      gin_file = (
          f'compiler_opt/rl/{problem}/gin_configs/{algorithm}_nn_agent.gin')
      gin_binding = ('generate_test_model.agent_config_type=' +
                     agent_config_type_dict[algorithm])
      test_setting.append((test_name, gin_file, gin_binding))

  return test_setting


class GenerateTestModelTest(parameterized.TestCase):

  @parameterized.named_parameters(*_get_test_settings())
  def test_generate_test_model(self, gin_file, gin_binding):

    tmp_dir = self.create_tempdir()

    with gin.unlock_config():
      gin.parse_config_files_and_bindings(
          config_files=[gin_file], bindings=[gin_binding], skip_unknown=True)

    with flagsaver.flagsaver(root_dir=tmp_dir.full_path):
      generate_test_model.generate_test_model()


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.env."""

import io
import contextlib
import ctypes
from unittest import mock
import subprocess

from typing import Dict, List, Optional

import tensorflow as tf
import numpy as np

from compiler_opt.rl import env
from compiler_opt.rl import corpus
from compiler_opt.rl import log_reader_test

_CLANG_PATH = '/test/clang/path'

_MOCK_MODULE = corpus.LoadedModuleSpec(
    name='module',
    loaded_ir=b'asdf',
    orig_options=('--opt_a', 'a', '--opt_b', 'b'),
)

_NUM_STEPS = 10


class MockTask(env.MLGOTask):
  """Implementation of mock task for testing."""

  def get_cmdline(self, clang_path: str, base_args: List[str],
                  interactive_base_path: Optional[str],
                  working_dir: str) -> List[str]:
    if interactive_base_path:
      interactive_args = [
          f'--interactive={interactive_base_path}',
      ]
    else:
      interactive_args = []
    return [clang_path] + base_args + interactive_args

  def get_module_scores(self, working_dir: str) -> Dict[str, float]:
    return {'default': 47}


# This mocks subprocess.Popen for interactive clang sessions
@contextlib.contextmanager
def mock_interactive_clang(cmdline, stderr, stdout):
  del stderr
  del stdout
  # do basic argument parsing
  fname = None
  for arg in cmdline:
    if arg.startswith('--interactive='):
      fname = arg[len('--interactive='):]
      break

  class MockProcess:

    def wait(self, timeout):
      pass

    def kill(self):
      pass

  if not fname:
    yield MockProcess()
    return
  # Create the fds for the pipes
  # (the env doesn't create the files, it assumes they are opened by clang)
  with io.FileIO(fname + '.out', 'wb+') as f_out:
    with io.FileIO(fname + '.in', 'rb+') as f_in:
      del f_in
      # Write the header describing the features/rewards
      f_out.write(
          log_reader_test.json_to_bytes({
              'features': [{
                  'name': 'times_called',
                  'port': 0,
                  'shape': [1],
                  'type': 'int64_t',
              },],
              'score': {
                  'name': 'reward',
                  'port': 0,
                  'shape': [1],
                  'type': 'float',
              },
          }))
      log_reader_test.write_nl(f_out)

      class MockInteractiveProcess(MockProcess):
        """Mock clang interactive process that writes the log."""

        def __init__(self):
          self._counter = 0

        # We poll the process at every call to get_observation to ensure the
        # clang process is still alive. So here, each time poll() is called,
        # write a new context
        def poll(self):
          if self._counter >= _NUM_STEPS:
            f_out.close()
            return None
          log_reader_test.write_context_marker(f_out,
                                               f'context_{self._counter}')
          log_reader_test.write_observation_marker(f_out, 0)
          log_reader_test.write_buff(f_out, [self._counter], ctypes.c_int64)
          log_reader_test.write_nl(f_out)
          log_reader_test.write_outcome_marker(f_out, 0)
          log_reader_test.write_buff(f_out, [3.14], ctypes.c_float)
          log_reader_test.write_nl(f_out)
          self._counter += 1
          return None

      yield MockInteractiveProcess()


class ClangSessionTest(tf.test.TestCase):

  @mock.patch('subprocess.Popen')
  def test_clang_session(self, mock_popen):
    mock_task = MockTask()
    with env.clang_session(
        _CLANG_PATH, _MOCK_MODULE, MockTask,
        interactive=False) as clang_session:
      del clang_session
      cmdline = mock_task.get_cmdline(_CLANG_PATH,
                                      list(_MOCK_MODULE.orig_options), None,
                                      '/tmp/mock/tmp/file')
      mock_popen.assert_called_once_with(
          cmdline, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

  @mock.patch('subprocess.Popen')
  def test_interactive_clang_session(self, mock_popen):
    mock_popen.side_effect = mock_interactive_clang

    with env.clang_session(
        _CLANG_PATH, _MOCK_MODULE, MockTask, interactive=True) as clang_session:
      for idx in range(_NUM_STEPS):
        obs = clang_session.get_observation()
        self.assertEqual(
            obs[env.OBS_KEY]['times_called'],
            np.array([idx], dtype=np.int64),
        )
        self.assertEqual(obs[env.CONTEXT_KEY], f'context_{idx}')
      mock_popen.assert_called_once()


class MLGOEnvironmentTest(tf.test.TestCase):

  @mock.patch('subprocess.Popen')
  def test_env(self, mock_popen):
    mock_popen.side_effect = mock_interactive_clang

    test_env = env.MLGOEnvironmentBase(
        clang_path=_CLANG_PATH,
        task_type=MockTask,
        obs_spec={},
        action_spec={},
    )

    for env_itr in range(3):
      del env_itr
      step = test_env.reset(_MOCK_MODULE)
      self.assertEqual(step[env.STEP_TYPE_KEY], env.FIRST_STEP_STR)

      for step_itr in range(_NUM_STEPS - 1):
        del step_itr
        step = test_env.step(np.array([1], dtype=np.int64))
        self.assertEqual(step[env.STEP_TYPE_KEY], env.MID_STEP_STR)

      step = test_env.step(np.array([1], dtype=np.int64))
      self.assertEqual(step[env.STEP_TYPE_KEY], env.LAST_STEP_STR)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extract IR for training.

Extract IR for training, either from a compile_commands.json file produced by
cmake, or a linker parameter list file.

Only run with
'python compiler_opt/tools/extract_ir.py ...'

The compilation is assumed to have been performed with clang, using
-fembed-bitcode=all passed to cc1 (i.e. pass clang -Xclang=-fembed-bitcode=all)

In a distributed ThinLTO case, the compilation is assumed to have been performed
specifying -mllvm -lto-embed-bitcode=post-merge-pre-opt.

In a local ThinLTO case, the compilation is assumedto have been performed
specifying -Wl,--save-temps=import -Wl,--thinlto-emit-index-files

To change the logging verbosity, pass an integer representing the desired
verbosity to the --verbosity flag. Use 0 for all logs, status information,
and detailed debug information, -1 for solely warnings, and -2 to not produce
any output.
"""

import json
import multiprocessing

from absl import app
from absl import flags
from absl import logging

from compiler_opt.tools import extract_ir_lib

flags.DEFINE_string(
    'input', None,
    'Input file or directory - either compile_commands.json, a linker parameter'
    'list, or a path to a directory containing object files.')
flags.DEFINE_enum(
    'input_type', 'json', ['json', 'params', 'directory'],
    'Input file type - json, params, or directory. params latter refers to lld'
    'params.')
flags.DEFINE_string('output_dir', None, 'Output directory')
flags.DEFINE_integer(
    'num_workers', None,
    'Number of parallel workers for objcopy. `None` for maximum available.')
flags.DEFINE_string('llvm_objcopy_path', 'llvm-objcopy', 'Path to llvm-objcopy')
flags.DEFINE_string(
    'obj_base_dir', '',
    'Base directory for object files. Defaults to current working dir.')
flags.DEFINE_string(
    'cmd_filter', None,
    'Include only those modules with a command line matching this regexp. '
    'Setting it to None for not filtering. Note that the regexp is applied '
    'independently for each separate command line option. For example, ^-Oz$ '
    'will match Oz - built binaries. Does not work with thinlto_build=lld.')
flags.DEFINE_enum(
    'thinlto_build', None, ['distributed', 'local'],
    'Set if the build was performed with either \'distributed\' or '
    '\'local\' ThinLTO. This ensures the thinlto.bc files are also copied. '
    'The build is assumed to have had '
    '-mllvm -lto-embed-bitcode=post-merge-pre-opt passed in the distributed '
    'case, or -Wl,--save-temps=import and -Wl,--thinlto-emit-index-files '
    'passed in the local case.')
flags.DEFINE_string(
    'cmd_section_name', '.llvmcmd',
    'The section name passed to llvm-objcopy. For ELF object files, the '
    'default .llvmcmd is correct. For Mach-O object files, one should use '
    'something like __LLVM,__cmdline')
flags.DEFINE_string(
    'bitcode_section_name', '.llvmbc',
    'The section name passed to llvm-objcopy. For ELF object files, the '
    'default .llvmbc is correct. For Mach-O object files, one should use '
    '__LLVM,__bitcode')

flags.mark_flag_as_required('output_dir')

FLAGS = flags.FLAGS


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  objs = []
  if FLAGS.input is not None and FLAGS.thinlto_build == 'local':
    raise ValueError('--thinlto_build=local cannot be run with --input')
  if FLAGS.input is None:
    if FLAGS.thinlto_build != 'local':
      raise ValueError('--input or --thinlto_build=local must be provided')
    objs = extract_ir_lib.load_for_lld_thinlto(FLAGS.obj_base_dir,
                                               FLAGS.output_dir)
  elif FLAGS.input_type == 'json':
    with open(FLAGS.input, encoding='utf-8') as f:
      objs = extract_ir_lib.load_from_compile_commands(
          json.load(f), FLAGS.output_dir)
  elif FLAGS.input_type == 'params':
    if not FLAGS.obj_base_dir:
      logging.info(
          '-obj_base_dir is unspecified, assuming current directory.'
          'If no objects are found, use this option to specify the root'
          'directory for the object file paths in the input file.')
    with open(FLAGS.input, encoding='utf-8') as f:
      objs = extract_ir_lib.load_from_lld_params(
          [l.strip() for l in f.readlines()], FLAGS.obj_base_dir,
          FLAGS.output_dir)
  elif FLAGS.input_type == 'directory':
    logging.warning(
        'Using the directory input is only recommended if the build system'
        'your project uses does not support any structured output that'
        'ml-compiler-opt understands. If your build system provides a'
        'structured compilation database, use that instead')
    objs = extract_ir_lib.load_from_directory(FLAGS.input, FLAGS.output_dir)
  else:
    logging.error('Unknown input type: %s', FLAGS.input_type)

  relative_output_paths = extract_ir_lib.run_extraction(
      objs, FLAGS.num_workers, FLAGS.llvm_objcopy_path, FLAGS.cmd_filter,
      FLAGS.thinlto_build, FLAGS.cmd_section_name, FLAGS.bitcode_section_name)

  extract_ir_lib.write_corpus_manifest(FLAGS.thinlto_build,
                                       relative_output_paths, FLAGS.output_dir)

  logging.info('Converted %d files out of %d',
               len(objs) - relative_output_paths.count(None), len(objs))


if __name__ == '__main__':
  multiprocessing.set_start_method('fork')
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class for coordinating blackbox optimization."""

import os
from absl import logging
import concurrent.futures
import dataclasses
import gin
import math
import numpy as np
import numpy.typing as npt
import tempfile
import tensorflow as tf
from typing import List, Optional, Protocol

from compiler_opt.distributed import buffered_scheduler
from compiler_opt.distributed.worker import FixedWorkerPool
from compiler_opt.es import blackbox_optimizers
from compiler_opt.es import policy_utils
from compiler_opt.rl import corpus
from compiler_opt.rl import policy_saver

# If less than 40% of requests succeed, skip the step.
_SKIP_STEP_SUCCESS_RATIO = 0.4


@gin.configurable
@dataclasses.dataclass(frozen=True)
class BlackboxLearnerConfig:
  """Hyperparameter configuration for BlackboxLearner."""

  # Total steps to train for
  total_steps: int

  # Name of the blackbox optimization algorithm
  blackbox_optimizer: blackbox_optimizers.Algorithm

  # What kind of ES training?
  #   - antithetic: for each perturbtation, try an antiperturbation
  #   - forward_fd: try total_num_perturbations independent perturbations
  est_type: blackbox_optimizers.EstimatorType

  # Should the rewards for blackbox optimization in a single step be normalized?
  fvalues_normalization: bool

  # How to update optimizer hyperparameters
  hyperparameters_update_method: blackbox_optimizers.UpdateMethod

  # Number of top performing perturbations to select in the optimizer
  # 0 means all
  num_top_directions: int

  # How many IR files to try a single perturbation on?
  num_ir_repeats_within_worker: int

  # How many times should we reuse IR to test different policies?
  num_ir_repeats_across_worker: int

  # How many IR files to sample from the test corpus at each iteration
  num_exact_evals: int

  # How many perturbations to attempt at each perturbation
  total_num_perturbations: int

  # How much to scale the stdev of the perturbations
  precision_parameter: float

  # Learning rate
  step_size: float


def _prune_skipped_perturbations(perturbations: List[npt.NDArray[np.float32]],
                                 rewards: List[Optional[float]]):
  """Remove perturbations that were skipped during the training step.

  Perturbations may be skipped due to an early exit condition or a server error
  (clang timeout, malformed training example, etc). The blackbox optimizer
  assumes that each perturbations has a valid reward, so we must remove any of
  these "skipped" perturbations.

  Pruning occurs in-place.

  Args:
    perturbations: the model perturbations used for the ES training step.
    rewards: the rewards for each perturbation.

  Returns:
    The number of perturbations that were pruned.
  """
  indices_to_prune = []
  for i, reward in enumerate(rewards):
    if reward is None:
      indices_to_prune.append(i)

  # Iterate in reverse so that the indices remain valid
  for i in reversed(indices_to_prune):
    del perturbations[i]
    del rewards[i]

  return len(indices_to_prune)


class PolicySaverCallableType(Protocol):
  """Protocol for the policy saver function.
  A Protocol is required to type annotate
  the function with keyword arguments"""

  def __call__(self, parameters: npt.NDArray[np.float32],
               policy_name: str) -> None:
    ...


class BlackboxLearner:
  """Implementation of blackbox learning."""

  def __init__(self,
               blackbox_opt: blackbox_optimizers.BlackboxOptimizer,
               sampler: corpus.Corpus,
               tf_policy_path: str,
               output_dir: str,
               policy_saver_fn: PolicySaverCallableType,
               model_weights: npt.NDArray[np.float32],
               config: BlackboxLearnerConfig,
               initial_step: int = 0,
               deadline: float = 30.0,
               seed: Optional[int] = None):
    """Construct a BlackboxLeaner.

    Args:
      blackbox_opt: the blackbox optimizer to use
      train_sampler: corpus_sampler for training data.
      tf_policy_path: where to write the tf policy
      output_dir: the directory to write all outputs
      policy_saver_fn: function to save a policy to cns
      model_weights: the weights of the current model
      config: configuration for blackbox optimization.
      stubs: grpc stubs to inlining/regalloc servers
      initial_step: the initial step for learning.
      deadline: the deadline in seconds for requests to the inlining server.
    """
    self._blackbox_opt = blackbox_opt
    self._sampler = sampler
    self._tf_policy_path = tf_policy_path
    self._output_dir = output_dir
    self._policy_saver_fn = policy_saver_fn
    self._model_weights = model_weights
    self._config = config
    self._step = initial_step
    self._deadline = deadline
    self._seed = seed

    # While we're waiting for the ES requests, we can
    # collect samples for the next round of training.
    self._samples = []

    self._summary_writer = tf.summary.create_file_writer(output_dir)

  def _get_perturbations(self) -> List[npt.NDArray[np.float32]]:
    """Get perturbations for the model weights."""
    perturbations = []
    rng = np.random.default_rng(seed=self._seed)
    for _ in range(self._config.total_num_perturbations):
      perturbations.append(
          rng.normal(size=(len(self._model_weights))) *
          self._config.precision_parameter)
    return perturbations

  def _get_rewards(
      self, results: List[concurrent.futures.Future]) -> List[Optional[float]]:
    """Convert ES results to reward numbers."""
    rewards = [None] * len(results)

    for i in range(len(results)):
      if not results[i].exception():
        rewards[i] = results[i].result()
      else:
        logging.info('Error retrieving result from future: %s',
                     str(results[i].exception()))

    return rewards

  def _update_model(self, perturbations: List[npt.NDArray[np.float32]],
                    rewards: List[float]) -> None:
    """Update the model given a list of perturbations and rewards."""
    self._model_weights = self._blackbox_opt.run_step(
        perturbations=np.array(perturbations),
        function_values=np.array(rewards),
        current_input=self._model_weights,
        current_value=np.mean(rewards))

  def _log_rewards(self, rewards: List[float]) -> None:
    """Log reward to console."""
    logging.info('Train reward: [%f]', np.mean(rewards))

  def _log_tf_summary(self, rewards: List[float]) -> None:
    """Log tensorboard data."""
    with self._summary_writer.as_default():
      tf.summary.scalar(
          'reward/average_reward_train', np.mean(rewards), step=self._step)

      tf.summary.histogram('reward/reward_train', rewards, step=self._step)

      train_regressions = [reward for reward in rewards if reward < 0]
      tf.summary.scalar(
          'reward/regression_probability_train',
          len(train_regressions) / len(rewards),
          step=self._step)

      tf.summary.scalar(
          'reward/regression_avg_train',
          np.mean(train_regressions) if len(train_regressions) > 0 else 0,
          step=self._step)

      # The "max regression" is the min value, as the regressions are negative.
      tf.summary.scalar(
          'reward/regression_max_train',
          min(train_regressions, default=0),
          step=self._step)

      train_wins = [reward for reward in rewards if reward > 0]
      tf.summary.scalar(
          'reward/win_probability_train',
          len(train_wins) / len(rewards),
          step=self._step)

  def _save_model(self) -> None:
    """Save the model."""
    logging.info('Saving the model.')
    self._policy_saver_fn(
        parameters=self._model_weights, policy_name=f'iteration{self._step}')

  def get_model_weights(self) -> npt.NDArray[np.float32]:
    return self._model_weights

  def _get_results(
      self, pool: FixedWorkerPool,
      perturbations: List[bytes]) -> List[concurrent.futures.Future]:
    if not self._samples:
      for _ in range(self._config.total_num_perturbations):
        sample = self._sampler.sample(self._config.num_ir_repeats_within_worker)
        self._samples.append(sample)
        # add copy of sample for antithetic perturbation pair
        if self._config.est_type == (
            blackbox_optimizers.EstimatorType.ANTITHETIC):
          self._samples.append(sample)

    compile_args = zip(perturbations, self._samples)

    _, futures = buffered_scheduler.schedule_on_worker_pool(
        action=lambda w, v: w.compile(v[0], v[1]),
        jobs=compile_args,
        worker_pool=pool)

    not_done = futures
    # wait for all futures to finish
    while not_done:
      # update lists as work gets done
      _, not_done = concurrent.futures.wait(
          not_done, return_when=concurrent.futures.FIRST_COMPLETED)

    return futures

  def _get_policy_as_bytes(self,
                           perturbation: npt.NDArray[np.float32]) -> bytes:
    sm = tf.saved_model.load(self._tf_policy_path)
    # devectorize the perturbation
    policy_utils.set_vectorized_parameters_for_policy(sm, perturbation)

    with tempfile.TemporaryDirectory() as tmpdir:
      sm_dir = os.path.join(tmpdir, 'sm')
      tf.saved_model.save(sm, sm_dir, signatures=sm.signatures)
      src = os.path.join(self._tf_policy_path, policy_saver.OUTPUT_SIGNATURE)
      dst = os.path.join(sm_dir, policy_saver.OUTPUT_SIGNATURE)
      tf.io.gfile.copy(src, dst)

      # convert to tflite
      tfl_dir = os.path.join(tmpdir, 'tfl')
      policy_saver.convert_mlgo_model(sm_dir, tfl_dir)

      # create and return policy
      policy_obj = policy_saver.Policy.from_filesystem(tfl_dir)
      return policy_obj.policy

  def run_step(self, pool: FixedWorkerPool) -> None:
    """Run a single step of blackbox learning.
    This does not instantaneously return due to several I/O
    and executions running while this waits for the responses"""
    logging.info('-' * 80)
    logging.info('Step [%d]', self._step)

    initial_perturbations = self._get_perturbations()
    # positive-negative pairs
    if self._config.est_type == blackbox_optimizers.EstimatorType.ANTITHETIC:
      initial_perturbations = [
          p for p in initial_perturbations for p in (p, -p)
      ]

    # convert to bytes for compile job
    # TODO: current conversion is inefficient.
    # consider doing this on the worker side
    perturbations_as_bytes = []
    for perturbation in initial_perturbations:
      perturbations_as_bytes.append(self._get_policy_as_bytes(perturbation))

    results = self._get_results(pool, perturbations_as_bytes)
    rewards = self._get_rewards(results)

    num_pruned = _prune_skipped_perturbations(initial_perturbations, rewards)
    logging.info('Pruned [%d]', num_pruned)
    min_num_rewards = math.ceil(_SKIP_STEP_SUCCESS_RATIO * len(results))
    if len(rewards) < min_num_rewards:
      logging.warning(
          'Skipping the step, too many requests failed: %d of %d '
          'train requests succeeded (required: %d)', len(rewards), len(results),
          min_num_rewards)
      return

    self._update_model(initial_perturbations, rewards)
    self._log_rewards(rewards)
    self._log_tf_summary(rewards)

    self._save_model()

    self._step += 1



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluate the current policy stored in reverb"""

import tempfile
import collections
import os
import time
from typing import List, Optional

from absl import logging

import tensorflow as tf
import tensorflow.compat.v1 as tf_v1
from tf_agents.experimental.distributed import reverb_variable_container
from tf_agents.train import learner
from tf_agents.train.utils import train_utils
from tf_agents.utils import common

from compiler_opt.rl import data_reader
from compiler_opt.rl import local_data_collector
from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import corpus
from compiler_opt.rl import agent_config
from compiler_opt.rl import registry
from compiler_opt.rl import policy_saver
from compiler_opt.rl import data_collector


def evaluate(root_dir: str, corpus_path: str,
             variable_container_server_address: str, num_workers: Optional[int],
             worker_manager_class):
  """Evaluate a given policy on the given corpus.

  Args:
    root_dir: path to write tensorboard summaries.
    corpus_path: path to the corpus to collect from.
    variable_container_server_address: address of the variable container in
      reverb.
    num_workers: number of workers in the pool to collect from.
    worker_manager_class: type of the worker manager to spawn the pool with.
  """

  # Set up the problem & initialize the agent
  logging.info('Initializing the distributed PPO agent')
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  agent_cfg = agent_config.DistributedPPOAgentConfig(
      time_step_spec=time_step_spec, action_spec=action_spec)
  agent = agent_config.create_agent(
      agent_cfg,
      preprocessing_layer_creator=problem_config
      .get_preprocessing_layer_creator())

  #policy = greedy_policy.GreedyPolicy(agent.policy)
  policy = agent.collect_policy
  policy_dict = {'policy': policy}

  # Initialize the reverb variable container
  logging.info('Connecting to the reverb server')
  train_step = tf_v1.train.get_or_create_global_step()
  model_id = common.create_variable('model_id')
  variables = {
      reverb_variable_container.POLICY_KEY: policy.variables(),
      reverb_variable_container.TRAIN_STEP_KEY: train_step,
      'model_id': model_id
  }
  variable_container = reverb_variable_container.ReverbVariableContainer(
      variable_container_server_address,
      table_names=[reverb_variable_container.DEFAULT_TABLE])
  variable_container.update(variables)

  # Setup the corpus
  logging.info('Constructing tf.data pipeline and module corpus')
  dataset_fn = data_reader.create_flat_sequence_example_dataset_fn(
      agent_cfg=agent_cfg)

  def sequence_example_iterator_fn(seq_ex: List[str]):
    return iter(dataset_fn(seq_ex).prefetch(tf.data.AUTOTUNE))

  cps = corpus.Corpus(
      data_path=corpus_path,
      additional_flags=problem_config.flags_to_add(),
      delete_flags=problem_config.flags_to_delete(),
      replace_flags=problem_config.flags_to_replace())

  summary_writer = tf.summary.create_file_writer(
      root_dir, flush_millis=5 * 1000)
  summary_writer.set_as_default()

  data_action_mean = tf.keras.metrics.Mean()
  data_reward_mean = tf.keras.metrics.Mean()

  # Run the experience collection loop.
  with worker_manager_class(
      worker_class=problem_config.get_runner_type(),
      count=num_workers,
      moving_average_decay_rate=1) as worker_pool:
    logging.info('constructed pool')
    collector = local_data_collector.LocalDataCollector(
        cps=cps,
        num_modules=128,
        worker_pool=worker_pool,
        parser=sequence_example_iterator_fn,
        reward_stat_map=collections.defaultdict(lambda: None),
        best_trajectory_repo=None)
    logging.info('constructed data_collector')
    last_step = -1
    rewards = []
    actions = []
    while True:
      with tempfile.TemporaryDirectory() as tmpdirname:
        saver = policy_saver.PolicySaver(policy_dict=policy_dict)
        saver.save(tmpdirname)
        policy_bytes = policy_saver.Policy.from_filesystem(
            os.path.join(tmpdirname, 'policy'))
      dataset_iter, monitor_dict = collector.collect_data(
          policy=policy_bytes, model_id=model_id.numpy())
      del monitor_dict

      for experience in dataset_iter:
        is_action = ~experience.is_boundary()

        # pylint: disable=not-callable
        data_action_mean.update_state(
            experience.action, sample_weight=is_action)
        # pylint: disable=not-callable
        data_reward_mean.update_state(
            experience.reward, sample_weight=is_action)

        if is_action:
          rewards.append(experience.reward)
          actions.append(experience.action)

      if last_step < train_step.numpy():
        last_step = train_step.numpy()
        with tf.name_scope('default/'):
          tf.summary.scalar(
              name='data_action_mean',
              # pylint: disable=not-callable
              data=data_action_mean.result(),
              step=train_step)
          tf.summary.scalar(
              name='data_reward_mean',
              # pylint: disable=not-callable
              data=data_reward_mean.result(),
              step=train_step)
          tf.summary.scalar(
              name='num_eval_experiences', data=len(rewards), step=train_step)
        tf.summary.histogram(name='reward', data=rewards, step=train_step)
        tf.summary.histogram(name='action', data=actions, step=train_step)
        with tf.name_scope('reward_distribution/'):
          reward_dist = data_collector.build_distribution_monitor(rewards)
          for k, v in reward_dist.items():
            tf.summary.scalar(name=k, data=v, step=train_step)
        data_action_mean.reset_state()
        data_reward_mean.reset_state()
        rewards = []
        actions = []

      variable_container.update(variables)
      logging.info('Evaluating with policy at step: %d', train_step.numpy())
      time.sleep(20)


def run_evaluate(root_dir: str, corpus_path: str,
                 variable_container_server_address: str,
                 num_workers: Optional[int], worker_manager_class):
  """Wait for the collect policy to be ready and run collect job."""
  # Wait for the collect policy to become available, then load it.
  policy_dir = os.path.join(root_dir, learner.POLICY_SAVED_MODEL_DIR,
                            learner.COLLECT_POLICY_SAVED_MODEL_DIR)
  policy = train_utils.wait_for_policy(policy_dir, load_specs_from_pbtxt=True)
  del policy

  evaluate(
      root_dir=root_dir,
      corpus_path=corpus_path,
      variable_container_server_address=variable_container_server_address,
      num_workers=num_workers,
      worker_manager_class=worker_manager_class)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library functions for making a corpus from arbitrary bitcode."""

import pathlib
import os
import shutil
import json

from typing import List, Optional

BITCODE_EXTENSION = '.bc'


def load_bitcode_from_directory(bitcode_base_dir: str) -> List[str]:
  """Finds bitcode files to extract from a given directory.

  Args:
    bitcode_base_dir: The base directory where the bitcode to be copied
      is from.
    output_dir: The directory to place the bitcode in.

  Returns an array of paths representing the relative path to the bitcode
  file from the base direcotry.
  """
  paths = [
      str(p)[:-len(BITCODE_EXTENSION)]
      for p in pathlib.Path(bitcode_base_dir).glob('**/*' + BITCODE_EXTENSION)
  ]

  return [
      os.path.relpath(full_path, start=bitcode_base_dir) for full_path in paths
  ]


def copy_bitcode(relative_paths: List[str], bitcode_base_dir: str,
                 output_dir: str) -> None:
  """Copies bitcode files from the base directory to the output directory.

  Args:
    relative_paths: An array of relative paths to bitcode files that are copied
      over to the output directory, preserving relative location.
    bitcode_base_dir: The base directory where the bitcode is located.
    output_dir: The output directory to place the bitcode in.
  """
  for relative_path in relative_paths:
    base_path = os.path.join(bitcode_base_dir,
                             relative_path + BITCODE_EXTENSION)
    destination_path = os.path.join(output_dir,
                                    relative_path + BITCODE_EXTENSION)
    os.makedirs(os.path.dirname(destination_path), exist_ok=True)
    shutil.copy(base_path, destination_path)


def write_corpus_manifest(relative_output_paths: List[str],
                          output_dir: str,
                          default_args: Optional[List[str]] = None) -> None:
  """Creates a corpus manifest describing the bitcode that has been found.

  Args:
    relative_output_paths: A list of paths to each bitcode file relative to the
      output directory.
    outout_dir: The output directory where the corpus is being created.
    default_args: An array of compiler flags that should be used to compile
      the bitcode when using further downstream tooling."""
  if default_args is None:
    default_args = []
  corpus_description = {
      'global_command_override': default_args,
      'has_thinlto': False,
      'modules': [path for path in relative_output_paths if path is not None]
  }

  with open(
      os.path.join(output_dir, 'corpus_description.json'),
      'w',
      encoding='utf-8') as description_file:
    json.dump(corpus_description, description_file, indent=2)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A utility for automatically setting up and running the LLVM test suite

This script will automatically do all of the setup in order to compile and run
the microbenchmarks in the llvm test suite. It will compile the requested model
into LLVM, compile the test suite using a two stage build to take advantage of
PGO, and then run all of the Microbenchmarks in the test suite, outputting the
data to a file of the user's choice.

Usage:
PYTHONPATH=$PYTHONPATH:. \
  python3 ./compiler_opt/benchmark/benchmark_llvm_test_suite.py \
  --advisor=release \
  --model_path=/tmp/model \
  --compile_llvm \
  --llvm_use_incremental \
  --compile_testsuite \
  --output_path=./output.json \
  --llvm_test_suite_path=/llvm-test-suite \
  --llvm_build_path=/llvm-build \
  --llvm_source_path=/llvm-project \
  --llvm_test_suite_build_path=/llvm-test-suite/build \
  --perf_counter=INSTRUCTIONS \
  --perf_counter=MEM_UOPS_RETIRED:ALL_LOADS \
  --perf_counter=MEM_UOPS_RETIRED:ALL_STORES

A lot of the flags listed above are just set to their default values. To see
what their default values are and whether or not you need to adjust them for
your setup, you can call the script with the --help command.
"""

import os
import shutil
import subprocess
import json

from absl import flags
from absl import app

from compiler_opt.benchmark import benchmarking_utils

default_tests = [
    'harris/harris', 'SLPVectorization/SLPVectorizationBenchmarks',
    'MemFunctions/MemFunctions',
    'LoopVectorization/LoopVectorizationBenchmarks',
    'LoopInterchange/LoopInterchange', 'LCALS/SubsetALambdaLoops/lcalsALambda',
    'LCALS/SubsetARawLoops/lcalsARaw', 'LCALS/SubsetBLambdaLoops/lcalsBLambda',
    'LCALS/SubsetBRawLoops/lcalsBRaw', 'LCALS/SubsetCLambdaLoops/lcalsCLambda',
    'LCALS/SubsetCRawLoops/lcalsCRaw',
    'ImageProcessing/AnisotropicDiffusion/AnisotropicDiffusion',
    'ImageProcessing/BilateralFiltering/BilateralFilter',
    'ImageProcessing/Blur/blur', 'ImageProcessing/Dilate/Dilate',
    'ImageProcessing/Dither/Dither',
    'ImageProcessing/Interpolation/Interpolation', 'Builtins/Int128/Builtins'
]

FLAGS = flags.FLAGS

flags.DEFINE_enum(
    'advisor', None, ['default', 'release'],
    'The regalloc advisor to be used for compiling'
    'the test suite')
flags.DEFINE_string('model_path', '',
                    'The path to the regalloc model for testing')
flags.DEFINE_boolean('compile_llvm', True,
                     'compiles llvm using the specified model path')
flags.DEFINE_boolean(
    'llvm_use_incremental', True,
    'recompile LLVM incrementally rather than doing a whole'
    'build')
flags.DEFINE_boolean(
    'compile_testsuite', True,
    'compiles the test suite using the specified advisor and'
    'model path')
flags.DEFINE_string('output_path', None,
                    'The output JSON file containing the test results')
flags.DEFINE_string('llvm_test_suite_path', '/llvm-test-suite',
                    'The path to the LLVM test suite repository root')
flags.DEFINE_string('llvm_build_path', '/llvm-build',
                    'The path to the llvm build')
flags.DEFINE_string('llvm_source_path', '/llvm-project',
                    'The path to the root of the llvm-project repositoy')
flags.DEFINE_string('llvm_test_suite_build_path', None,
                    'The path to the llvm test suite build')
flags.DEFINE_multi_string(
    'tests_to_run', default_tests,
    'Tests compiled with google benchmark to run,'
    'with paths from the base of the ./microbenchmarks'
    'directory')
flags.DEFINE_multi_string(
    'perf_counter', [], 'A perf counter to be used (may be defined more than'
    'once).')

flags.mark_flag_as_required('output_path')


def build_test_suite(regalloc_advisor: str, llvm_test_suite_build_path: str,
                     llvm_build_path: str, llvm_test_suite_source_path: str):
  """Builds the LLVM test suite using the specified regalloc advisor

  This function just builds the llvm test suite from scratch. The only two
  advisor modes are release and default as the main function automatically
  compiles the advisor to be evaluated as the release mode model into llvm.
  This script then compiles the test suite using PGO so that benchmarks will
  return valid results.

  Args:
    regalloc_advisor: The regalloc advisor (default or release) to use when
      compiling
    llvm_test_suite_build_path: the path to place the llvm test suite build in
    llvm_build_path: The directory to where the LLVM/clang build has been
      completed
  """
  llvm_c_compiler_path = os.path.join(llvm_build_path, './bin/clang')
  llvm_cxx_compiler_path = os.path.join(llvm_build_path, './bin/clang++')
  llvm_lit_path = os.path.join(llvm_build_path, './bin/llvm-lit')

  if os.path.exists(llvm_test_suite_build_path):
    shutil.rmtree(llvm_test_suite_build_path)

  os.makedirs(llvm_test_suite_build_path)

  cpp_flags = f'-mllvm -regalloc-enable-advisor={regalloc_advisor}'

  cmake_config_command_stage_1 = [
      'cmake', '-G', 'Ninja', '-DTEST_SUITE_PROFILE_GENERATE=ON',
      '-DTEST_SUITE_RUN_TYPE=train',
      f'-DCMAKE_C_COMPILER={llvm_c_compiler_path}',
      f'-DCMAKE_CXX_COMPILER={llvm_cxx_compiler_path}',
      f'-DCMAKE_CXX_FLAGS=\'{cpp_flags}\'', f'-DCMAKE_C_FLAGS=\'{cpp_flags}\'',
      '-DBENCHMARK_ENABLE_LIBPFM=ON', '-DTEST_SUITE_BENCHMARKING_ONLY=ON',
      llvm_test_suite_source_path
  ]

  with subprocess.Popen(
      cmake_config_command_stage_1,
      cwd=llvm_test_suite_build_path) as cmake_stage_1_process:
    cmake_stage_1_process.wait()

  cmake_compile_command = ['cmake', '--build', '.']
  with subprocess.Popen(
      cmake_compile_command,
      cwd=llvm_test_suite_build_path) as cmake_stage_1_build_process:
    cmake_stage_1_build_process.wait()

  lit_test_runner_command = [f'{llvm_lit_path}', '.']

  with subprocess.Popen(
      lit_test_runner_command,
      cwd=llvm_test_suite_build_path) as lit_test_runner_process:
    lit_test_runner_process.wait()

  cmake_config_command_stage_2 = [
      'cmake', '-G', 'Ninja', '-DTEST_SUITE_PROFILE_GENERATE=OFF',
      '-DTEST_SUITE_PROFILE_USE=ON', '-DTEST_SUITE_RUN_TYPE=ref', '.'
  ]

  with subprocess.Popen(
      cmake_config_command_stage_2,
      cwd=llvm_test_suite_build_path) as cmake_stage_2_process:
    cmake_stage_2_process.wait()

  with subprocess.Popen(
      cmake_compile_command,
      cwd=llvm_test_suite_build_path) as cmake_stage_2_build_process:
    cmake_stage_2_build_process.wait()


def main(_):
  if FLAGS.llvm_test_suite_build_path is None:
    FLAGS.llvm_test_suite_build_path = os.path.join(FLAGS.llvm_test_suite_path,
                                                    './build')
  if FLAGS.compile_llvm:
    benchmarking_utils.build_llvm(FLAGS.model_path, FLAGS.llvm_use_incremental,
                                  FLAGS.llvm_build_path, FLAGS.llvm_source_path)
  if FLAGS.compile_testsuite:
    build_test_suite(FLAGS.advisor, FLAGS.llvm_test_suite_build_path,
                     FLAGS.llvm_build_path, FLAGS.llvm_test_suite_path)
  completed_benchmarks = []
  for test in FLAGS.tests_to_run:
    test_path = os.path.join(FLAGS.llvm_test_suite_build_path,
                             './MicroBenchmarks/' + test)
    completed_benchmarks.extend(
        benchmarking_utils.run_microbenchmark(test_path, FLAGS.perf_counter))
  with open(FLAGS.output_path, 'w', encoding='UTF-8') as output_file:
    json_to_write = {'benchmarks': completed_benchmarks}
    output_file.write(json.dumps(json_to_write, indent=4))


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.compilation_runner."""

import os
import string
import subprocess
import threading
import time
from unittest import mock

from absl import flags
import tensorflow as tf

# This is https://github.com/google/pytype/issues/764
from google.protobuf import text_format  # pytype: disable=pyi-error

from compiler_opt.rl import compilation_runner
from compiler_opt.rl import constant
from compiler_opt.rl import corpus
from compiler_opt.rl import policy_saver

_DEFAULT_FEATURE_VALUE = 12
_POLICY_FEATURE_VALUE = 34

_DEFAULT_REWARD = 10
_POLICY_REWARD = 8

_MOVING_AVERAGE_DECAY_RATE = 0.8


def _get_sequence_example_with_reward(feature_value, reward):
  sequence_example_text = string.Template("""
    feature_lists {
      feature_list {
        key: "feature_0"
        value {
          feature { int64_list { value: $feature_value } }
          feature { int64_list { value: $feature_value } }
        }
      }
      feature_list {
        key: "reward"
        value {
          feature { float_list { value: $reward } }
          feature { float_list { value: $reward } }
        }
      }
    }""").substitute(
      feature_value=feature_value, reward=reward)
  return text_format.Parse(sequence_example_text, tf.train.SequenceExample())


def _get_sequence_example(feature_value):
  sequence_example_text = string.Template("""
    feature_lists {
      feature_list {
        key: "feature_0"
        value {
          feature { int64_list { value: $feature_value } }
          feature { int64_list { value: $feature_value } }
        }
      }
    }""").substitute(feature_value=feature_value)
  return text_format.Parse(sequence_example_text, tf.train.SequenceExample())


def _mock_compile_fn(file_paths, tf_policy_path, reward_only, workdir):  # pylint: disable=unused-argument
  del file_paths
  if tf_policy_path:
    sequence_example = _get_sequence_example(_POLICY_FEATURE_VALUE)
    native_size = _POLICY_REWARD
  else:
    sequence_example = _get_sequence_example(_DEFAULT_FEATURE_VALUE)
    native_size = _DEFAULT_REWARD

  if reward_only:
    return {'default': (None, native_size)}
  else:
    return {'default': (sequence_example, native_size)}


_mock_policy = policy_saver.Policy(bytes(), bytes())

_mock_loaded_module_spec = corpus.LoadedModuleSpec(
    name='dummy', loaded_ir=bytes())


class CompilationRunnerTest(tf.test.TestCase):

  def assertListProtoEqual(self, a, b):
    self.assertEqual(len(a), len(b))
    a.sort()
    b.sort()
    for (x, y) in zip(a, b):
      self.assertProtoEquals(x, y)

  @mock.patch(constant.BASE_MODULE_DIR +
              '.compilation_runner.CompilationRunner.compile_fn')
  def test_policy(self, mock_compile_fn):
    mock_compile_fn.side_effect = _mock_compile_fn
    runner = compilation_runner.CompilationRunner(
        moving_average_decay_rate=_MOVING_AVERAGE_DECAY_RATE)
    data = runner.collect_data(
        loaded_module_spec=_mock_loaded_module_spec, policy=_mock_policy)
    self.assertEqual(2, mock_compile_fn.call_count)

    expected_example = _get_sequence_example_with_reward(
        _POLICY_FEATURE_VALUE, 0.1998)
    self.assertListProtoEqual([expected_example], [
        tf.train.SequenceExample.FromString(x)
        for x in data.serialized_sequence_examples
    ])
    self.assertEqual(2, data.length)
    self.assertCountEqual(
        {
            'default':
                compilation_runner.RewardStat(
                    default_reward=_DEFAULT_REWARD,
                    moving_average_reward=_DEFAULT_REWARD *
                    _MOVING_AVERAGE_DECAY_RATE + _POLICY_REWARD *
                    (1 - _MOVING_AVERAGE_DECAY_RATE))
        }, data.reward_stats)
    self.assertAllClose([0.1998002], data.rewards)
    self.assertAllClose([8], data.policy_rewards)

  @mock.patch(constant.BASE_MODULE_DIR +
              '.compilation_runner.CompilationRunner.compile_fn')
  def test_default(self, mock_compile_fn):
    mock_compile_fn.side_effect = _mock_compile_fn
    runner = compilation_runner.CompilationRunner(
        moving_average_decay_rate=_MOVING_AVERAGE_DECAY_RATE)

    data = runner.collect_data(loaded_module_spec=_mock_loaded_module_spec)
    # One call when we ask for the default policy, because it can provide both
    # trace and default size.
    self.assertEqual(1, mock_compile_fn.call_count)

    expected_example = _get_sequence_example_with_reward(
        _DEFAULT_FEATURE_VALUE, 0)
    self.assertListProtoEqual([expected_example], [
        tf.train.SequenceExample.FromString(x)
        for x in data.serialized_sequence_examples
    ])
    self.assertEqual(2, data.length)
    self.assertCountEqual(
        {
            'default':
                compilation_runner.RewardStat(
                    default_reward=_DEFAULT_REWARD,
                    moving_average_reward=_DEFAULT_REWARD)
        }, data.reward_stats)
    self.assertAllClose([0], data.rewards)
    self.assertAllClose([10], data.policy_rewards)

  @mock.patch(constant.BASE_MODULE_DIR +
              '.compilation_runner.CompilationRunner.compile_fn')
  def test_given_default_size(self, mock_compile_fn):
    mock_compile_fn.side_effect = _mock_compile_fn
    runner = compilation_runner.CompilationRunner(
        moving_average_decay_rate=_MOVING_AVERAGE_DECAY_RATE)

    data = runner.collect_data(
        loaded_module_spec=_mock_loaded_module_spec,
        policy=_mock_policy,
        reward_stat={
            'default':
                compilation_runner.RewardStat(
                    default_reward=_DEFAULT_REWARD, moving_average_reward=7)
        })
    self.assertEqual(1, mock_compile_fn.call_count)

    expected_example = _get_sequence_example_with_reward(
        _POLICY_FEATURE_VALUE,
        1 - (_POLICY_REWARD + constant.DELTA) / (7 + constant.DELTA))
    self.assertListProtoEqual([expected_example], [
        tf.train.SequenceExample.FromString(x)
        for x in data.serialized_sequence_examples
    ])
    self.assertEqual(2, data.length)
    self.assertCountEqual(
        {
            'default':
                compilation_runner.RewardStat(
                    default_reward=_DEFAULT_REWARD,
                    moving_average_reward=7 * _MOVING_AVERAGE_DECAY_RATE +
                    _POLICY_REWARD * (1 - _MOVING_AVERAGE_DECAY_RATE))
        }, data.reward_stats)
    self.assertAllClose([0.199800], data.rewards)
    self.assertAllClose([8], data.policy_rewards)

  @mock.patch(constant.BASE_MODULE_DIR +
              '.compilation_runner.CompilationRunner.compile_fn')
  def test_exception_handling(self, mock_compile_fn):
    mock_compile_fn.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd='error')
    runner = compilation_runner.CompilationRunner(
        moving_average_decay_rate=_MOVING_AVERAGE_DECAY_RATE)

    with self.assertRaisesRegex(subprocess.CalledProcessError, 'error'):
      _ = runner.collect_data(
          loaded_module_spec=_mock_loaded_module_spec,
          policy=_mock_policy,
          reward_stat=None)
    self.assertEqual(1, mock_compile_fn.call_count)

  def test_start_subprocess_output(self):
    cm = compilation_runner.WorkerCancellationManager()
    output = compilation_runner.start_cancellable_process(
        ['ls', '-l'], timeout=100, cancellation_manager=cm, want_output=True)
    if output:
      output_str = output.decode('utf-8')
    else:
      self.fail('output should have been non-empty')
    self.assertNotEmpty(output_str)

  def test_timeout_kills_process(self):
    sentinel_file = os.path.join(flags.FLAGS.test_tmpdir,
                                 'test_timeout_kills_test_file')
    if os.path.exists(sentinel_file):
      os.remove(sentinel_file)
    with self.assertRaises(subprocess.TimeoutExpired):
      compilation_runner.start_cancellable_process(
          ['bash', '-c', 'sleep 1s ; touch ' + sentinel_file],
          timeout=0.5,
          cancellation_manager=None)
    time.sleep(2)
    self.assertFalse(os.path.exists(sentinel_file))

  def test_pause_resume(self):
    cm = compilation_runner.WorkerCancellationManager()
    start_time = time.time()

    def stop_and_start():
      time.sleep(0.25)
      cm.pause_all_processes()
      time.sleep(1)
      cm.resume_all_processes()

    threading.Thread(target=stop_and_start).start()
    compilation_runner.start_cancellable_process(['sleep', '0.5'],
                                                 30,
                                                 cancellation_manager=cm)
    # should be at least 1 second due to the pause.
    self.assertGreater(time.time() - start_time, 1)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with test executables using gtest (ie Chromium)"""

import sys
import subprocess
import re

from joblib import Parallel, delayed
from typing import Tuple, List, Optional, Dict
from absl import logging


def run_test(test_executable: str, test_name: str, perf_counters: List[str]):
  """Runs a specific test

  This function executes a specific test in a gtest executable using
  the --gtest_filter flag to select a specific test. All performance
  counter information is captured using perf stat.

  Args:
    test_executable: the path to the gtest executable containing the test
    test_name: the name of the test to be run
    perf_counters: A string list of performance counters recognized by
      perf stat (platform dependent)
  """
  command_vector = ['perf', 'stat']
  for perf_counter in perf_counters:
    command_vector.extend(['-e', perf_counter])
  command_vector.extend([test_executable, f'--gtest_filter={test_name}'])
  with subprocess.Popen(
      command_vector, stdout=subprocess.PIPE,
      stderr=subprocess.PIPE) as process:
    out, err = process.communicate()
    decoded_stderr = err.decode('UTF-8')
    decoded_stdout = out.decode('UTF-8')
    if process.returncode != 0:
      logging.warning('test %s failed', test_name)
      raise RuntimeError(f'Test executable failed while running {test_name}')
    elif 'PASSED' not in decoded_stdout:
      logging.warning('test %s does not exist', test_name)
      raise RuntimeError(f'No test {test_name} exists in test executable')
    # all of the output from perf stat is on STDERR
    return decoded_stderr


def parse_perf_stat_output(perf_stat_output: str, perf_counters: List[str]):
  """Parses raw output from perf stat

  This function takes in the raw decoded output from perf stat
  and parses it into a dictionary containing each of the requested
  performance counters as a key.

  Args:
    perf_stat_output: raw decoded output from a perf stat run
    perf_counters: A list of strings of valid perf stat
      performance counters
  """
  counters_dict = {}
  for line in perf_stat_output.split('\n'):
    for perf_counter in perf_counters:
      if perf_counter in line:
        count_string = re.findall(r'^\s*\d*', line)[0].replace(' ', '')
        count = int(count_string)
        counters_dict[perf_counter] = count
  return counters_dict


def run_and_parse(test_description: Tuple[str, str, List[str]]):
  """Runs a test and processes the output of an individual test

  This function takes in a description of an individual test, runs the test
  to get the perf stat output, and then returns the parsed perf stat output
  in the form of a dictionary

  Args:
    test_description: a tuple in the form of (executable path, test name,
      performance counters to collect) that describes the test
  """
  test_executable, test_name, performance_counters = test_description
  try:
    test_output = run_test(test_executable, test_name, performance_counters)
    print(f'Finished running test {test_name}', file=sys.stderr)
    return (test_name, parse_perf_stat_output(test_output,
                                              performance_counters))
  except RuntimeError:
    return None


def run_test_suite(test_suite_description: Dict[str, List[str]],
                   test_executable: str, perf_counters: List[str],
                   num_threads: Optional[int]):
  """Runs an entire test suite

  This function takes in a test set description in the form of a path to a JSON
  file and runs all of the tests within that test suite description, capturing
  all of the performance counters requested. This function also allows different
  tests to be run in parallel for non-parallel sensitive performance counters
  such as ones tracking overall loads and stores, but parallelism should be used
  with extreme caution while benchmarking.

  Args:
    test_suite_description: A python dictionary containing an array with the
      key tests which has all the tests to run
    test_executable: A path to the gtest executable being described by the test
      description JSON
    perf_counters: A list of strings of valid perf performance counters that
      are to be collected.
    num_threads: The number of threads to use when running tests. Set to 1 by
      default. Be very cautious about running benchmarks in parallel.
  """

  if num_threads is None:
    num_threads = 1

  test_descriptions = []
  for test in test_suite_description['tests']:
    test_descriptions.append((test_executable, test, perf_counters))

  test_data_output = Parallel(n_jobs=num_threads)(
      delayed(run_and_parse)(test_description)
      for test_description in test_descriptions)

  formatted_test_data = []
  for test_instance in test_data_output:
    if test_instance:
      test_info = {'name': test_instance[0], 'iterations': 1}
      test_info.update(test_instance[1])
      formatted_test_data.append(test_info)

  return formatted_test_data


def get_gtest_testlist_raw(path_to_executable: str):
  """Gets raw output of a gtest executable's test list

  Takes in a path to a gtest executable and uses the flag --gtest_list_tests
  to get a list of all the tests and then returns the raw output so that it
  can later be parsed.

  Args:
    path_to_executable: A path to the gtest executable for which a test list
      is desired
  """
  command_vector = [path_to_executable, '--gtest_list_tests']
  with subprocess.Popen(
      command_vector, stdout=subprocess.PIPE,
      stderr=subprocess.PIPE) as process:
    out = process.communicate()[0]
    return out.decode('UTF-8')


def parse_gtest_tests(gtest_output_raw: str):
  """Parses gtest test list output into a Python list

  Loops through each line in the raw output obtained from the
  get_gtest_testlist_raw functions and relies on the current output
  structure in order to be able to parse the test names.

  Args:
    gtest_output_raw: A string containing the decoded
      raw output from a gtest executable run with the
      --gtest_list_tests flag
  """
  test_list = []
  current_test_prefix = ''
  gtest_output_split = gtest_output_raw.split('\n')
  current_index = 0
  # skip to the actual test list
  while current_index < len(gtest_output_split):
    current_string = gtest_output_split[current_index]
    test_matches = re.findall(r'^[a-zA-Z]*\.$', current_string)
    if len(test_matches) != 0:
      break
    current_index += 1
  while current_index < len(gtest_output_split):
    current_string = gtest_output_split[current_index]
    if len(current_string) == 0:
      current_index += 1
      continue
    # get the test name
    test_match = re.findall(r'^\s*\S*', current_string)[0].replace(' ', '')
    if test_match[len(test_match) - 1] == '.':
      # We've found a new prefix
      current_test_prefix = test_match
      current_index += 1
      continue
    test_list.append(current_test_prefix + test_match)
    current_index += 1
  return test_list



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for generate_default_trace."""
import json
import os
from unittest import mock

from absl import flags
from absl.testing import absltest
from absl.testing import flagsaver
import gin
import tensorflow as tf

# This is https://github.com/google/pytype/issues/764
from google.protobuf import text_format  # pytype: disable=pyi-error
from compiler_opt.rl import compilation_runner
from compiler_opt.tools import generate_default_trace

from tf_agents.system import system_multiprocessing as multiprocessing

flags.FLAGS['num_workers'].allow_override = True
flags.FLAGS['gin_files'].allow_override = True
flags.FLAGS['gin_bindings'].allow_override = True


@gin.configurable(module='runners')
class MockCompilationRunner(compilation_runner.CompilationRunner):
  """A compilation runner just for test."""

  def __init__(self, sentinel=None):
    assert sentinel == 42
    super().__init__()

  def collect_data(self,
                   loaded_module_spec,
                   policy=None,
                   reward_stat=None,
                   model_id=None):
    sequence_example_text = """
      feature_lists {
        feature_list {
          key: "feature_0"
          value {
            feature { int64_list { value: 1 } }
            feature { int64_list { value: 1 } }
          }
        }
      }"""
    sequence_example = text_format.Parse(sequence_example_text,
                                         tf.train.SequenceExample())

    return compilation_runner.CompilationResult(
        sequence_examples=[sequence_example],
        reward_stats={
            'default':
                compilation_runner.RewardStat(
                    default_reward=1, moving_average_reward=2)
        },
        rewards=[1.2],
        policy_rewards=[18],
        keys=['default'],
        model_id=model_id)


class GenerateDefaultTraceTest(absltest.TestCase):

  def setUp(self):
    with gin.unlock_config():
      gin.parse_config_files_and_bindings(
          config_files=['compiler_opt/rl/inlining/gin_configs/common.gin'],
          bindings=['runners.MockCompilationRunner.sentinel=42'])
    return super().setUp()

  @mock.patch('compiler_opt.rl.inlining.InliningConfig.get_runner_type')
  def test_generate_trace(self, mock_get_runner):

    tmp_dir = self.create_tempdir()
    module_names = ['a', 'b', 'c', 'd']

    with tf.io.gfile.GFile(
        os.path.join(tmp_dir.full_path, 'corpus_description.json'), 'w') as f:
      json.dump({'modules': module_names, 'has_thinlto': False}, f)

    for module_name in module_names:
      with tf.io.gfile.GFile(
          os.path.join(tmp_dir.full_path, module_name + '.bc'), 'w') as f:
        f.write(module_name)

      with tf.io.gfile.GFile(
          os.path.join(tmp_dir.full_path, module_name + '.cmd'), 'w') as f:
        f.write('-cc1')

    mock_compilation_runner = MockCompilationRunner
    mock_get_runner.return_value = mock_compilation_runner

    with flagsaver.flagsaver(
        data_path=tmp_dir.full_path,
        num_workers=2,
        output_path=os.path.join(tmp_dir.full_path, 'output'),
        output_performance_path=os.path.join(tmp_dir.full_path,
                                             'output_performance'),
    ):
      generate_default_trace.generate_trace()


if __name__ == '__main__':
  multiprocessing.handle_main(absltest.main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
#
#
# This is a port of the code by Krzysztof Choromanski, Deepali Jain and Vikas
# Sindhwani, based on the portfolio of Blackbox optimization algorithms listed
# below:
#
# "On Blackbox Backpropagation and Jacobian Sensing"; K. Choromanski,
#  V. Sindhwani, NeurIPS 2017
# "Optimizing Simulations with Noise-Tolerant Structured Exploration"; K.
#  Choromanski, A. Iscen, V. Sindhwani, J. Tan, E. Coumans, ICRA 2018
# "Structured Evolution with Compact Architectures for Scalable Policy
#  Optimization"; K. Choromanski, M. Rowland, V. Sindhwani, R. Turner, A.
#  Weller, ICML 2018, https://arxiv.org/abs/1804.02395
#  "From Complexity to Simplicity: Adaptive ES-Active Subspaces for Blackbox
#   Optimization";  K. Choromanski, A. Pacchiano, J. Parker-Holder, Y. Tang, V.
#   Sindhwani, NeurIPS 2019
# "i-Sim2Real: Reinforcement Learning on Robotic Policies in Tight Human-Robot
#  Interaction Loops"; L. Graesser, D. D'Ambrosio, A. Singh, A. Bewley, D. Jain,
#  K. Choromanski, P. Sanketi , CoRL 2022, https://arxiv.org/abs/2207.06572
# "Agile Catching with Whole-Body MPC and Blackbox Policy Learning"; S.
#  Abeyruwan, A. Bewley, N. Boffi, K. Choromanski, D. D'Ambrosio, D. Jain, P.
#  Sanketi, A. Shankar, V. Sindhwani, S. Singh, J. Slotine, S. Tu, L4DC,
#  https://arxiv.org/abs/2306.08205
# "Robotic Table Tennis: A Case Study into a High Speed Learning System"; A.
#  Bewley, A. Shankar, A. Iscen, A. Singh, C. Lynch, D. D'Ambrosio, D. Jain,
#  E. Coumans, G. Versom, G. Kouretas, J. Abelian, J. Boyd, K. Oslund,
#  K. Reymann, K. Choromanski, L. Graesser, M. Ahn, N. Jaitly, N. Lazic,
#  P. Sanketi, P. Xu, P. Sermanet, R. Mahjourian, S. Abeyruwan, S. Kataoka,
#  S. Moore, T. Nguyen, T. Ding, V. Sindhwani, V. Vanhoucke, W. Gao, Y. Kuang,
#  to be presented at RSS 2023
###############################################################################
r"""Tests for blackbox_optimization_algorithms."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from compiler_opt.es import blackbox_optimizers
from compiler_opt.es.blackbox_optimizers import EstimatorType, UpdateMethod, RegressionType
from compiler_opt.es import gradient_ascent_optimization_algorithms

perturbation_array = np.array([[0, 1], [2, -1], [4, 2],
                               [-2, -2], [0, 3], [0, -3], [0, 4], [0, -4],
                               [-1, 5], [1, -5], [2, 6], [8, -6]])
function_value_array = np.array(
    [-1, 1, 10, -10, -2, 2, -0.5, 0.5, 4, -4, -8, 8])


class BlackboxOptimizationAlgorithmsTest(parameterized.TestCase):

  @parameterized.parameters(
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 3,
       np.array([[4, 2], [2, 6], [-1, 5], [-2, -2], [8, -6], [1, -5]
                ]), np.array([10, -8, 4, -10, 8, -4])),
      (perturbation_array, function_value_array, EstimatorType.FORWARD_FD, 5,
       np.array([[4, 2], [8, -6], [-1, 5], [0, -3], [2, -1]
                ]), np.array([10, 8, 4, 2, 1])))
  def test_filtering(self, perturbations, function_values, est_type,
                     num_top_directions, expected_ps, expected_fs):
    top_ps, top_fs = blackbox_optimizers.filter_top_directions(
        perturbations, function_values, est_type, num_top_directions)
    np.testing.assert_array_equal(expected_ps, top_ps)
    np.testing.assert_array_equal(expected_fs, top_fs)

  @parameterized.parameters(
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 3,
       np.array([100, -16])), (perturbation_array, function_value_array,
                               EstimatorType.FORWARD_FD, 5, np.array([76, -9])),
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 0,
       np.array([102, -34])),
      (perturbation_array, function_value_array, EstimatorType.FORWARD_FD, 0,
       np.array([74, -34])))
  def test_monte_carlo_gradient(self, perturbations, function_values, est_type,
                                num_top_directions, expected_gradient):
    precision_parameter = 0.1
    step_size = 0.01
    current_value = 2
    blackbox_object = blackbox_optimizers.MonteCarloBlackboxOptimizer(
        precision_parameter, est_type, False, UpdateMethod.NO_METHOD, None,
        step_size, num_top_directions)
    current_input = np.zeros(2)
    step = blackbox_object.run_step(perturbations, function_values,
                                    current_input, current_value)
    gradient = step * (precision_parameter**2) / step_size
    if num_top_directions == 0:
      gradient *= len(perturbations)
    else:
      gradient *= num_top_directions

    np.testing.assert_array_almost_equal(expected_gradient, gradient)

  @parameterized.parameters(
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 3,
       np.array([100, -16])), (perturbation_array, function_value_array,
                               EstimatorType.FORWARD_FD, 5, np.array([76, -9])),
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 0,
       np.array([102, -34])),
      (perturbation_array, function_value_array, EstimatorType.FORWARD_FD, 0,
       np.array([74, -34])))
  def test_monte_carlo_gradient_with_gradient_ascent_optimizer(
      self, perturbations, function_values, est_type, num_top_directions,
      expected_gradient):
    precision_parameter = 0.1
    step_size = 0.01
    current_value = 2
    gradient_ascent_optimizer = (
        gradient_ascent_optimization_algorithms.MomentumOptimizer(
            step_size, 0.0))
    blackbox_object = (
        blackbox_optimizers.MonteCarloBlackboxOptimizer(
            precision_parameter, est_type, False, UpdateMethod.NO_METHOD, None,
            None, num_top_directions, gradient_ascent_optimizer))
    current_input = np.zeros(2)
    step = blackbox_object.run_step(perturbations, function_values,
                                    current_input, current_value)
    gradient = step * (precision_parameter**2) / step_size
    if num_top_directions == 0:
      gradient *= len(perturbations)
    else:
      gradient *= num_top_directions

    np.testing.assert_array_almost_equal(expected_gradient, gradient)

  @parameterized.parameters(
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 3,
       np.array([0.00483, -0.007534])),
      (perturbation_array, function_value_array, EstimatorType.FORWARD_FD, 5,
       np.array([0.012585, 0.000748])),
      (perturbation_array, function_value_array, EstimatorType.ANTITHETIC, 0,
       np.array([0.019319, -0.030134])),
      (perturbation_array, function_value_array, EstimatorType.FORWARD_FD, 0,
       np.array([0.030203, 0.001796])))
  def test_sklearn_gradient(self, perturbations, function_values, est_type,
                            num_top_directions, expected_gradient):
    precision_parameter = 0.1
    step_size = 0.01
    current_value = 2
    regularizer = 0.01
    gradient_ascent_optimizer = (
        gradient_ascent_optimization_algorithms.MomentumOptimizer(
            step_size, 0.0))
    blackbox_object = blackbox_optimizers.SklearnRegressionBlackboxOptimizer(
        RegressionType.RIDGE, regularizer, est_type, True,
        UpdateMethod.NO_METHOD, [], None, gradient_ascent_optimizer)
    current_input = np.zeros(2)
    step = blackbox_object.run_step(perturbations, function_values,
                                    current_input, current_value)
    gradient = step * (precision_parameter**2) / step_size
    if num_top_directions == 0:
      gradient *= len(perturbations)
    else:
      gradient *= num_top_directions

    np.testing.assert_array_almost_equal(expected_gradient, gradient)


class SecondorderBlackboxOptimizersTest(absltest.TestCase):

  def setUp(self):
    """Create common data matrices for tests.

    We will use a particular quadratic problem
    f(x) = 1/2x^TAx + b^Tx + c
    in several of the tests and initialize it here.
    The matrix A is indefinite and has eigs
    [2.15, 0.53, -2.67]
    """
    super().setUp()
    # pylint: disable=bad-whitespace,invalid-name
    self.A = np.array([[1, -1, 0], [-1, 0, 2], [0, 2, -1]])
    self.b = np.array([1, 0, 1])
    self.c = 2.5
    self.multiply_Av = lambda v: np.matmul(self.A, v)
    # pylint: enable=bad-whitespace,invalid-name

  def testQuadraticModelFunctionValue(self):
    quad_model = blackbox_optimizers.QuadraticModel(self.multiply_Av, self.b,
                                                    self.c)
    x = np.array([1, 0, -2])
    self.assertEqual(0.0, quad_model.f(x))

  def testQuadraticModelGradient(self):
    quad_model = blackbox_optimizers.QuadraticModel(self.multiply_Av, self.b,
                                                    self.c)
    x = np.array([1, 0, -2])
    gradient = quad_model.grad(x)
    self.assertEqual(gradient[0], 2.0)
    self.assertEqual(gradient[1], -5.0)
    self.assertEqual(gradient[2], 3.0)

  def testMakeProjector(self):
    radius = 3.0
    x = np.array([3.0, 3.0])
    projector = blackbox_optimizers.make_projector(radius)
    projected_point = projector(x)
    self.assertTrue(np.isclose(3.0 / np.sqrt(2), projected_point[0]))
    self.assertTrue(np.isclose(3.0 / np.sqrt(2), projected_point[0]))

  def testProjectedGradientOptimizer(self):
    """Simple test for correctness of PGD.

    Minimize the function
    (x + 1)^2 + (y - 1)^2
    over the nonnegative orthant.
    The exact solution is (0,1).
    """

    class GenericFunction(blackbox_optimizers.QuadraticModel):

      def __init__(self):  # pylint: disable=super-init-not-called
        pass

      def f(self, x):
        return (x[0] + 1)**2 + (x[1] - 1)**2

      def grad(self, x):
        return np.array([2 * (x[0] + 1), 2 * (x[1] - 1)])

    def projector(x):
      return np.maximum(0, x)

    objective_function = GenericFunction()
    pgd_params = {'const_step_size': 0.5}
    pgd_optimizer = blackbox_optimizers.ProjectedGradientOptimizer(
        objective_function, projector, pgd_params, np.array([1, 1]))
    while (pgd_optimizer.get_iterations() <= 10 and
           pgd_optimizer.get_x_diff_norm() > 1e-9):
      pgd_optimizer.run_step()
    solution = pgd_optimizer.get_solution()
    self.assertLess(abs(solution[0]), 1e-8)
    self.assertLess(abs(solution[1] - 1.0), 1e-8)

  def testTrustRegionSubproblem(self):
    """Simple test for trust region subproblem solver with indefinite quadratic.

    Test the Trust Region Subproblem by solving the quadratic
    1/2x^TAx + b^Tx over the ball of radius 2, where A, b are created
    in setUp.
    """
    quad_model = blackbox_optimizers.QuadraticModel(self.multiply_Av, self.b,
                                                    self.c)
    tr_params = {
        'radius': 2,
        'problem_dim': 3,
        'subproblem_solver': 'PGD',
        'subproblem_maxiter': 100,
        'sub_terminate_stable': 1e-9
    }
    tr_subproblem_solver = blackbox_optimizers.TrustRegionSubproblemOptimizer(
        quad_model, tr_params)
    tr_subproblem_solver.solve_trust_region_subproblem()
    solution = tr_subproblem_solver.get_solution()
    self.assertLess(abs(solution[0] - 0.03157944), 1e-8)
    self.assertLess(abs(solution[1] - 1.12525306), 1e-8)
    self.assertLess(abs(solution[2] + 1.65312076), 1e-8)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tool for making a corpus from arbitrary bitcode.

To create a corpus from a set of bitcode files in an input directory, run
the following command:

PYTHONPATH=$PYTHONPATH:. python3 ./compiler_opt/tools/make_corpus.py \
  --input_dir=<path to input directory> \
  --output_dir=<path to output directory> \
  --default_args="<list of space separated flags>"
"""

from absl import app
from absl import flags
from absl import logging

from compiler_opt.tools import make_corpus_lib

flags.DEFINE_string('input_dir', None, 'The input directory.')
flags.DEFINE_string('output_dir', None, 'The output directory.')
flags.DEFINE_string(
    'default_args', '',
    'The compiler flags to compile with when using downstream tooling.')

flags.mark_flag_as_required('input_dir')
flags.mark_flag_as_required('output_dir')

FLAGS = flags.FLAGS


def main(_):
  logging.warning(
      'Using this tool does not guarantee that the bitcode is taken at '
      'the correct stage for consumption during model training. Make '
      'sure to validate assumptions about where the bitcode is coming '
      'from before using it in production.')
  relative_paths = make_corpus_lib.load_bitcode_from_directory(FLAGS.input_dir)
  make_corpus_lib.copy_bitcode(relative_paths, FLAGS.input_dir,
                               FLAGS.output_dir)
  make_corpus_lib.write_corpus_manifest(relative_paths, FLAGS.output_dir,
                                        FLAGS.default_args.split())


if __name__ == '__main__':
  app.run(main)
